r"""
=========================
 Estimation API
=========================

"""

__docformat__ = "restructuredtext en"

import warnings

import numpy as np
from scipy.sparse import coo_matrix
from scipy.sparse import csr_matrix
from scipy.sparse import issparse
from scipy.sparse.sputils import isdense

from deeptime.util.types import ensure_dtraj_list
from . import dense
from . import sparse

__author__ = "Benjamin Trendelkamp-Schroer, Martin Scherer, Frank Noe"
__copyright__ = "Copyright 2014, Computational Molecular Biology Group, FU-Berlin"
__credits__ = ["Benjamin Trendelkamp-Schroer", "Martin Scherer", "Fabian Paul", "Frank Noe"]

__version__ = "2.0.0"
__maintainer__ = "Martin Scherer"
__email__ = "m.scherer AT fu-berlin DOT de"

__all__ = ['bootstrap_trajectories',
           'bootstrap_counts',
           'count_matrix',
           'connected_sets',
           'effective_count_matrix',
           'error_perturbation',
           'is_connected',
           'largest_connected_set',
           'largest_connected_submatrix',
           'log_likelihood',
           'prior_const',
           'prior_neighbor',
           'prior_rev',
           'rate_matrix',
           'sample_tmatrix',
           'tmatrix_cov',
           'tmatrix_sampler',
           'transition_matrix'
           ]


################################################################################
# Count matrix
################################################################################

def count_matrix(dtraj, lag, sliding=True, sparse_return=True, nstates=None):
    r"""Generate a count matrix from given microstate trajectory. :footcite:`prinz2011markov`

    Parameters
    ----------
    dtraj : array_like or list of array_like
        Discretized trajectory or list of discretized trajectories
    lag : int
        Lagtime in trajectory steps
    sliding : bool, optional
        If true the sliding window approach
        is used for transition counting.
    sparse_return : bool (optional)
        Whether to return a dense or a sparse matrix.
    nstates : int, optional
        Enforce a count-matrix with shape=(nstates, nstates)

    Returns
    -------
    C : scipy.sparse.coo_matrix
        The count matrix at given lag in coordinate list format.

    Notes
    -----
    Transition counts can be obtained from microstate trajectory using
    two methods. Couning at lag and slidingwindow counting.

    **Lag**

    This approach will skip all points in the trajectory that are
    seperated form the last point by less than the given lagtime
    :math:`\tau`.

    Transition counts :math:`c_{ij}(\tau)` are generated according to

    .. math:: c_{ij}(\tau) = \sum_{k=0}^{\left \lfloor \frac{N}{\tau} \right \rfloor -2}
                                        \chi_{i}(X_{k\tau})\chi_{j}(X_{(k+1)\tau}).

    :math:`\chi_{i}(x)` is the indicator function of :math:`i`, i.e
    :math:`\chi_{i}(x)=1` for :math:`x=i` and :math:`\chi_{i}(x)=0` for
    :math:`x \neq i`.

    **Sliding**

    The sliding approach slides along the trajectory and counts all
    transitions sperated by the lagtime :math:`\tau`.

    Transition counts :math:`c_{ij}(\tau)` are generated according to

    .. math:: c_{ij}(\tau)=\sum_{k=0}^{N-\tau-1} \chi_{i}(X_{k}) \chi_{j}(X_{k+\tau}).

    References
    ----------
    .. footbibliography::

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.estimation import count_matrix

    >>> dtraj = np.array([0, 0, 1, 0, 1, 1, 0])
    >>> tau = 2

    Use the sliding approach first

    >>> C_sliding = count_matrix(dtraj, tau)

    The generated matrix is a sparse matrix in CSR-format. For
    convenient printing we convert it to a dense ndarray.

    >>> C_sliding.toarray()
    array([[1., 2.],
           [1., 1.]])

    Let us compare to the count-matrix we obtain using the lag
    approach

    >>> C_lag = count_matrix(dtraj, tau, sliding=False)
    >>> C_lag.toarray()
    array([[0., 1.],
           [1., 1.]])

    """
    # convert dtraj input, if it contains out of nested python lists to
    # a list of int ndarrays.
    dtraj = ensure_dtraj_list(dtraj)
    return sparse.count_matrix.count_matrix_coo2_mult(dtraj, lag, sliding=sliding,
                                                      sparse=sparse_return, nstates=nstates)


def effective_count_matrix(dtrajs, lag, average='row', mact=1.0, n_jobs=None, callback=None):
    r""" Computes the statistically effective transition count matrix

    Given a list of discrete trajectories, compute the effective number of statistically uncorrelated transition
    counts at the given lag time. First computes the full sliding-window counts :math:`c_{ij}(tau)`. Then uses
    :func:`statistical_inefficiencies` to compute statistical inefficiencies :math:`I_{ij}(tau)`. The number of
    effective counts in a row is then computed as

    .. math:
        c_i^{\mathrm{eff}}(tau) = \sum_j I_{ij}(tau) c_{ij}(tau)

    and the effective transition counts are obtained by scaling the rows accordingly:

    .. math:
        c_{ij}^{\mathrm{eff}}(tau) = \frac{c_i^{\mathrm{eff}}(tau)}{c_i(tau)} c_{ij}(tau)

    This procedure is not yet published, but a manuscript is in preparation [1]_.

    Parameters
    ----------
    dtrajs : list of int-iterables
        discrete trajectories
    lag : int
        lag time
    average : str, default='row'
        Use either of 'row', 'all', 'none', with the following consequences:
        'none': the statistical inefficiency is applied separately to each
            transition count (not recommended)
        'row': the statistical inefficiency is averaged (weighted) by row
            (recommended).
        'all': the statistical inefficiency is averaged (weighted) over all
            transition counts (not recommended).
    mact : float, default=1.0
        multiplier for the autocorrelation time. We tend to underestimate the
        autocorrelation time (and thus overestimate effective counts)
        because the autocorrelation function is truncated when it passes
        through 0 in order to avoid numerical instabilities.
        This is a purely heuristic factor trying to compensate this effect.
        This parameter might be removed in the future when a more robust
        estimation method of the autocorrelation time is used.
    n_jobs : int, default=None
        If None, uses all available logical cores, otherwise the function will be evaluated with as
        many processes as specified (must then be positive).
    callback : callable, default=None
        will be called for every statistical inefficiency computed (number of nonzero elements in count matrix).
        If n_jobs is greater one, the callback will be invoked per finished batch.

    See also
    --------
    statistical_inefficiencies
        is used for computing the statistical inefficiencies of sliding window transition counts

    References
    ----------
    .. [1] Noe, F. and H. Wu: in preparation (2015)

    """
    from deeptime.util.parallel import handle_n_jobs
    n_jobs = handle_n_jobs(n_jobs)
    dtrajs = ensure_dtraj_list(dtrajs)
    return sparse.effective_counts.effective_count_matrix(dtrajs, lag, average=average, mact=mact, n_jobs=n_jobs,
                                                          callback=callback)


################################################################################
# Bootstrapping data
################################################################################

def bootstrap_trajectories(trajs, correlation_length):
    r"""Generates a randomly resampled trajectory segments.

    Parameters
    ----------
    trajs : array-like or array-like of array-like
        single or multiple trajectories. Every trajectory is assumed to be
        a statistically independent realization. Note that this is often not true and
        is a weakness with the present bootstrapping approach.

    correlation_length : int
        Correlation length (also known as the or statistical inefficiency) of the data.
        If set to < 1 or > L, where L is the longest trajectory length, the
        bootstrapping will sample full trajectories.
        We suggest to select the largest implied timescale or relaxation timescale as a
        conservative estimate of the correlation length. If this timescale is unknown,
        it's suggested to use full trajectories (set timescale to < 1) or come up with
        a rough estimate. For computing the error on specific observables, one may use
        shorter timescales, because the relevant correlation length is the integral of
        the autocorrelation function of the observables of interest [3]_. The slowest
        implied timescale is an upper bound for that correlation length, and therefore
        a conservative estimate [4]_.

    Notes
    -----
    This function can be called multiple times in order to generate randomly
    resampled trajectory data. In order to compute error bars on your observable
    of interest, call this function to generate resampled trajectories, and
    put them into your estimator. The standard deviation of such a sample of
    the observable is a model for the standard error.

    Implements a moving block bootstrapping procedure [1]_ for generation of
    randomly resampled count matrixes from discrete trajectories. The corrlation length
    determines the size of trajectory blocks that will remain contiguous.
    For a single trajectory N with correlation length t_corr < N,
    we will sample floor(N/t_corr) subtrajectories of length t_corr using starting time t.
    t is a uniform random number in [0, N-t_corr-1].
    When multiple trajectories are available, N is the total number of timesteps
    over all trajectories, the algorithm will generate resampled data with a total number
    of N (or slightly larger) time steps. Each trajectory of length n_i has a probability
    of n_i to be selected. Trajectories of length n_i <= t_corr are returned completely.
    For longer trajectories, segments of length t_corr are randomly generated.

    Note that like all error models for correlated time series data, Bootstrapping
    just gives you a model for the error given a number of assumptions [2]_. The most
    critical decisions are: (1) is this approach meaningful at all (only if the
    trajectories are statistically independent realizations), and (2) select
    an appropriate timescale of the correlation length (see below).
    Note that transition matrix sampling from the Dirichlet distribution is a
    much better option from a theoretical point of view, but may also be
    computationally more demanding.

    References
    ----------
    .. [1] H. R. Kuensch. The jackknife and the bootstrap for general
        stationary observations, Ann. Stat. 3, 1217-41 (1989).
    .. [2] B. Efron. Bootstrap methods: Another look at the jackknife.
        Ann. Statist. 7 1-26 (1979).
    .. [3] T.W. Anderson. The Statistical Analysis of Time Series
        Wiley, New York (1971).
    .. [4] F. Noe and F. Nueske: A variational approach to modeling
        slow processes in stochastic dynamical systems.  SIAM
        Multiscale Model. Simul., 11 . pp. 635-655 (2013)

    """
    return dense.bootstrapping.bootstrap_trajectories(trajs, correlation_length)


def bootstrap_counts(dtrajs, lagtime, corrlength=None):
    r"""Generates a randomly resampled count matrix given the input coordinates.

    Parameters
    ----------
    dtrajs : array-like or array-like of array-like
        single or multiple discrete trajectories. Every trajectory is assumed to be
        a statistically independent realization. Note that this is often not true and
        is a weakness with the present bootstrapping approach.

    lagtime : int
        the lag time at which the count matrix will be evaluated

    corrlength : int, optional, default=None
        the correlation length of the discrete trajectory. N / corrlength counts will be generated,
        where N is the total number of frames. If set to None (default), corrlength = lagtime will be used.

    Notes
    -----
    This function can be called multiple times in order to generate randomly
    resampled realizations of count matrices. For each of these realizations
    you can estimate a transition matrix, and from each of them computing the
    observables of your interest. The standard deviation of such a sample of
    the observable is a model for the standard error.

    The bootstrap will be generated by sampling N/corrlength counts at time tuples (t, t+lagtime),
    where t is uniformly sampled over all trajectory time frames in [0,n_i-lagtime].
    Here, n_i is the length of trajectory i and N = sum_i n_i is the total number of frames.

    See also
    --------
    bootstrap_trajectories

    """
    dtrajs = ensure_dtraj_list(dtrajs)
    return dense.bootstrapping.bootstrap_counts(dtrajs, lagtime, corrlength=corrlength)


################################################################################
# Connectivity
################################################################################

def connected_sets(C, directed=True):
    r"""Compute connected sets of microstates.

    Connected components for a directed graph with edge-weights
    given by the count matrix.

    Parameters
    ----------
    C : scipy.sparse matrix
        Count matrix specifying edge weights.
    directed : bool, optional
       Whether to compute connected components for a directed  or
       undirected graph. Default is True.

    Returns
    -------
    cc : list of arrays of integers
        Each entry is an array containing all vertices (states) in the
        corresponding connected component. The list is sorted
        according to the size of the individual components. The
        largest connected set is the first entry in the list, lcc=cc[0].

    Notes
    -----
    Viewing the count matrix as the adjacency matrix of a (directed) graph
    the connected components are given by the connected components of that
    graph. Connected components of a graph can be efficiently computed
    using Tarjan's algorithm :footcite:`tarjan1972depth`.

    References
    ----------
    .. footbibliography::

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.estimation import connected_sets

    >>> C = np.array([[10, 1, 0], [2, 0, 3], [0, 0, 4]])
    >>> cc_directed = connected_sets(C)
    >>> cc_directed
    [array([0, 1]), array([2])]

    >>> cc_undirected = connected_sets(C, directed=False)
    >>> cc_undirected
    [array([0, 1, 2])]

    """
    if isdense(C):
        return sparse.connectivity.connected_sets(csr_matrix(C), directed=directed)
    else:
        return sparse.connectivity.connected_sets(C, directed=directed)


def largest_connected_set(C, directed=True):
    r"""Largest connected component for a directed graph with edge-weights
    given by the count matrix.

    Parameters
    ----------
    C : scipy.sparse matrix
        Count matrix specifying edge weights.
    directed : bool, optional
       Whether to compute connected components for a directed  or
       undirected graph. Default is True.

    Returns
    -------
    lcc : array of integers
        The largest connected component of the directed graph.

    See also
    --------
    connected_sets

    Notes
    -----
    Viewing the count matrix as the adjacency matrix of a (directed)
    graph the largest connected set is the largest connected set of
    nodes of the corresponding graph. The largest connected set of a graph
    can be efficiently computed using Tarjan's algorithm :footcite:`tarjan1972depth`.

    References
    ----------
    .. footbibliography::

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.estimation import largest_connected_set

    >>> C =  np.array([[10, 1, 0], [2, 0, 3], [0, 0, 4]])
    >>> lcc_directed = largest_connected_set(C)
    >>> lcc_directed
    array([0, 1])

    >>> lcc_undirected = largest_connected_set(C, directed=False)
    >>> lcc_undirected
    array([0, 1, 2])

    """
    if isdense(C):
        return sparse.connectivity.largest_connected_set(csr_matrix(C), directed=directed)
    else:
        return sparse.connectivity.largest_connected_set(C, directed=directed)


def largest_connected_submatrix(C, directed=True, lcc=None):
    r"""Compute the count matrix on the largest connected set.

    Parameters
    ----------
    C : scipy.sparse matrix
        Count matrix specifying edge weights.
    directed : bool, optional
       Whether to compute connected components for a directed or
       undirected graph. Default is True
    lcc : (M,) ndarray, optional
       The largest connected set

    Returns
    -------
    C_cc : scipy.sparse matrix
        Count matrix of largest completely
        connected set of vertices (states)

    See also
    --------
    largest_connected_set

    Notes
    -----
    Viewing the count matrix as the adjacency matrix of a (directed)
    graph the larest connected submatrix is the adjacency matrix of
    the largest connected set of the corresponding graph. The largest
    connected submatrix can be efficiently computed using Tarjan's algorithm :footcite:`tarjan1972depth`.

    References
    ----------
    .. footbibliography::

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.estimation import largest_connected_submatrix

    >>> C = np.array([[10, 1, 0], [2, 0, 3], [0, 0, 4]])

    >>> C_cc_directed = largest_connected_submatrix(C)
    >>> C_cc_directed # doctest: +ELLIPSIS
    array([[10,  1],
           [ 2,  0]]...)

    >>> C_cc_undirected = largest_connected_submatrix(C, directed=False)
    >>> C_cc_undirected # doctest: +ELLIPSIS
    array([[10,  1,  0],
           [ 2,  0,  3],
           [ 0,  0,  4]]...)

    """
    if isdense(C):
        return sparse.connectivity.largest_connected_submatrix(csr_matrix(C), directed=directed, lcc=lcc).toarray()
    else:
        return sparse.connectivity.largest_connected_submatrix(C, directed=directed, lcc=lcc)


def is_connected(C, directed=True):
    """Check connectivity of the given matrix.

    Parameters
    ----------
    C : scipy.sparse matrix
        Count matrix specifying edge weights.
    directed : bool, optional
       Whether to compute connected components for a directed or
       undirected graph. Default is True.

    Returns
    -------
    is_connected: bool
        True if C is connected, False otherwise.

    See also
    --------
    largest_connected_submatrix

    Notes
    -----
    A count matrix is connected if the graph having the count matrix
    as adjacency matrix has a single connected component. Connectivity
    of a graph can be efficiently checked using Tarjan's algorithm :footcite:`tarjan1972depth`.

    References
    ----------
    .. footbibliography::

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.estimation import is_connected

    >>> C = np.array([[10, 1, 0], [2, 0, 3], [0, 0, 4]])
    >>> is_connected(C)
    False

    >>> is_connected(C, directed=False)
    True

    """
    if isdense(C):
        return sparse.connectivity.is_connected(csr_matrix(C), directed=directed)
    else:
        return sparse.connectivity.is_connected(C, directed=directed)


################################################################################
# priors
################################################################################

def prior_neighbor(C, alpha=0.001):
    r"""Neighbor prior for the given count matrix.

    Parameters
    ----------
    C : (M, M) ndarray or scipy.sparse matrix
        Count matrix
    alpha : float (optional)
        Value of prior counts

    Returns
    -------
    B : (M, M) ndarray or scipy.sparse matrix
        Prior count matrix

    Notes
    ------
    The neighbor prior :math:`b_{ij}` is defined as

    .. math:: b_{ij}=\left \{ \begin{array}{rl}
                     \alpha & c_{ij}+c_{ji}>0 \\
                     0      & \text{else}
                     \end{array} \right .

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.estimation import prior_neighbor

    >>> C = np.array([[10, 1, 0], [2, 0, 3], [0, 1, 4]])
    >>> B = prior_neighbor(C)
    >>> B
    array([[0.001, 0.001, 0.   ],
           [0.001, 0.   , 0.001],
           [0.   , 0.001, 0.001]])

    """

    if isdense(C):
        B = sparse.prior.prior_neighbor(csr_matrix(C), alpha=alpha)
        return B.toarray()
    else:
        return sparse.prior.prior_neighbor(C, alpha=alpha)


def prior_const(C, alpha=0.001):
    r"""Constant prior for given count matrix.

    Parameters
    ----------
    C : (M, M) ndarray or scipy.sparse matrix
        Count matrix
    alpha : float (optional)
        Value of prior counts

    Returns
    -------
    B : (M, M) ndarray
        Prior count matrix

    Notes
    -----
    The prior is defined as

    .. math:: \begin{array}{rl} b_{ij}= \alpha & \forall i, j \end{array}

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.estimation import prior_const

    >>> C = np.array([[10, 1, 0], [2, 0, 3], [0, 1, 4]])
    >>> B = prior_const(C)
    >>> B
    array([[0.001, 0.001, 0.001],
           [0.001, 0.001, 0.001],
           [0.001, 0.001, 0.001]])

    """
    if isdense(C):
        return sparse.prior.prior_const(C, alpha=alpha)
    else:
        warnings.warn("Prior will be a dense matrix for sparse input")
        return sparse.prior.prior_const(C, alpha=alpha)


__all__.append('prior_const')


def prior_rev(C, alpha=-1.0):
    r"""Prior counts for sampling of reversible transition
    matrices.

    Prior is defined as

    b_ij= alpha if i<=j
    b_ij=0         else

    Parameters
    ----------
    C : (M, M) ndarray or scipy.sparse matrix
        Count matrix
    alpha : float (optional)
        Value of prior counts

    Returns
    -------
    B : (M, M) ndarray
        Matrix of prior counts

    Notes
    -----
    The reversible prior is a matrix with -1 on the upper triangle.
    Adding this prior respects the fact that
    for a reversible transition matrix the degrees of freedom
    correspond essentially to the upper triangular part of the matrix.

    The prior is defined as

    .. math:: b_{ij} = \left \{ \begin{array}{rl}
                       \alpha & i \leq j \\
                       0      & \text{elsewhere}
                       \end{array} \right .

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.estimation import prior_rev

    >>> C = np.array([[10, 1, 0], [2, 0, 3], [0, 1, 4]])
    >>> B = prior_rev(C)
    >>> B
    array([[-1., -1., -1.],
           [ 0., -1., -1.],
           [ 0.,  0., -1.]])

    """
    if isdense(C):
        return sparse.prior.prior_rev(C, alpha=alpha)
    else:
        # warnings.warn("Prior will be a dense matrix for sparse input")
        return sparse.prior.prior_rev(C, alpha=alpha)


################################################################################
# Transition matrix
################################################################################

def transition_matrix(C, reversible=False, mu=None, method='auto',
                      maxiter: int = 1000000, maxerr: float = 1e-8,
                      rev_pisym : bool = False, return_statdist: bool = False, warn_not_converged: bool = True,
                      sparse_newton: bool = False):
    r"""Estimate the transition matrix from the given countmatrix. :footcite:`prinz2011markov`
    :footcite:`bowman2009progress` :footcite:`trendelkamp2015estimation`

    Parameters
    ----------
    C : numpy ndarray or scipy.sparse matrix
        Count matrix
    reversible : bool (optional)
        If True restrict the ensemble of transition matrices
        to those having a detailed balance symmetry otherwise
        the likelihood optimization is carried out over the whole
        space of stochastic matrices.
    mu : array_like
        The stationary distribution of the MLE transition matrix.
    method : str
        Select which implementation to use for the estimation.
        One of 'auto', 'dense' and 'sparse', optional, default='auto'.
        'dense' always selects the dense implementation, 'sparse' always selects
        the sparse one.
        'auto' selects the most efficient implementation according to
        the sparsity structure of the matrix: if the occupation of the C
        matrix is less then one third, select sparse. Else select dense.
        The type of the T matrix returned always matches the type of the
        C matrix, irrespective of the method that was used to compute it.

    Other Parameters
    ----------------
    maxiter : int, optional, default=1000000
        Optional parameter with reversible = True.
        maximum number of iterations before the method exits
    maxerr : float, optional, default=1e-8
        Optional parameter with reversible = True.
        convergence tolerance for transition matrix estimation.
        This specifies the maximum change of the Euclidean norm of relative
        stationary probabilities (:math:`x_i = \sum_k x_{ik}`). The relative stationary probability changes
        :math:`e_i = (x_i^{(1)} - x_i^{(2)})/(x_i^{(1)} + x_i^{(2)})` are used in order to track changes in small
        probabilities. The Euclidean norm of the change vector, :math:`|e_i|_2`, is compared to maxerr.
    rev_pisym : bool, optional, default=False
        Fast computation of reversible transition matrix by normalizing
        :math:`x_{ij} = \pi_i p_{ij} + \pi_j p_{ji}`. :math:`p_{ij}` is the direct
        (nonreversible) estimate and :math:`pi_i` is its stationary distribution.
        This estimator is asympotically unbiased but not maximum likelihood.
    return_statdist : bool, optional, default=False
        Optional parameter with reversible = True.
        If set to true, the stationary distribution is also returned
    return_conv : bool, optional, default=False
        Optional parameter with reversible = True.
        If set to true, the likelihood history and the pi_change history is returned.
    warn_not_converged : bool, optional, default=True
        Prints a warning if not converged.
    sparse_newton : bool, optional, default=False
        If True, use the experimental primal-dual interior-point solver for sparse input/computation method.

    Returns
    -------
    result : Union[array_like, Tuple[array_like, np.ndarray]]
       The MLE transition matrix and optionally the stationary distribution if `return_statist=True`.
       The transition matrix has the same data type (dense or sparse) as the input matrix C.

    Notes
    -----
    The transition matrix is a maximum likelihood estimate (MLE) of
    the probability distribution of transition matrices with
    parameters given by the count matrix.

    References
    ----------
    .. footbibliography::

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.estimation import transition_matrix

    >>> C = np.array([[10, 1, 1], [2, 0, 3], [0, 1, 4]])

    Non-reversible estimate

    >>> T_nrev = transition_matrix(C)
    >>> print(np.array_str(T_nrev, precision=3))
    [[0.833 0.083 0.083]
     [0.4   0.    0.6  ]
     [0.    0.2   0.8  ]]

    Reversible estimate

    >>> T_rev = transition_matrix(C, reversible=True)
    >>> print(np.array_str(T_rev, precision=3))
    [[0.833 0.104 0.063]
     [0.351 0.    0.649]
     [0.049 0.151 0.8  ]]

    Reversible estimate with given stationary vector

    >>> mu = np.array([0.7, 0.01, 0.29])
    >>> T_mu = transition_matrix(C, reversible=True, mu=mu)
    >>> print(np.array_str(T_mu, precision=3))
    [[0.948 0.006 0.046]
     [0.429 0.    0.571]
     [0.111 0.02  0.869]]
    """
    if issparse(C):
        sparse_input_type = True
    elif isdense(C):
        sparse_input_type = False
    else:
        raise NotImplementedError('C has an unknown type.')

    if method == 'dense':
        sparse_computation = False
    elif method == 'sparse':
        sparse_computation = True
    elif method == 'auto':
        # heuristically determine whether is't more efficient to do a dense of sparse computation
        if sparse_input_type:
            dof = C.getnnz()
        else:
            dof = np.count_nonzero(C)
        dimension = C.shape[0]
        if dimension * dimension < 3 * dof:
            sparse_computation = False
        else:
            sparse_computation = True
    else:
        raise ValueError(('method="%s" is no valid choice. It should be one of'
                          '"dense", "sparse" or "auto".') % method)

    # convert input type
    if sparse_computation and not sparse_input_type:
        C = coo_matrix(C)
    if not sparse_computation and sparse_input_type:
        C = C.toarray()

    if reversible:
        if mu is None:
            if sparse_computation:
                if rev_pisym:
                    result = sparse.transition_matrix.transition_matrix_reversible_pisym(
                        C, return_statdist=return_statdist
                    )
                elif sparse_newton:
                    from .sparse.mle.newton.mle_rev import solve_mle_rev
                    result = solve_mle_rev(C, tol=maxerr, maxiter=maxiter, return_statdist=return_statdist)
                else:
                    result = sparse.mle.mle_trev(C, maxerr=maxerr, maxiter=maxiter,
                                                 warn_not_converged=warn_not_converged,
                                                 return_statdist=return_statdist)
            else:
                if rev_pisym:
                    result = dense.transition_matrix.transition_matrix_reversible_pisym(
                        C, return_statdist=return_statdist
                    )
                else:
                    result = dense.mle.mle_trev(C, maxerr=maxerr, maxiter=maxiter,
                                                warn_not_converged=warn_not_converged, return_statdist=return_statdist)
        else:
            if sparse_computation:
                # Sparse, reversible, fixed pi (currently using dense with sparse conversion)
                result = sparse.mle.mle_trev_given_pi(C, mu, maxerr=maxerr, maxiter=maxiter,
                                                      warn_not_converged=warn_not_converged)
            else:
                result = dense.mle.mle_trev_given_pi(C, mu, maxerr=maxerr, maxiter=maxiter)
    else:  # nonreversible estimation
        if mu is None:
            if sparse_computation:
                # Sparse,  nonreversible
                result = sparse.transition_matrix.transition_matrix_non_reversible(C)
            else:
                # Dense,  nonreversible
                result = dense.transition_matrix.transition_matrix_non_reversible(C)
            # Both methods currently do not have an iterate of pi, so we compute it here for consistency.
            if return_statdist:
                from ..analysis import stationary_distribution
                mu = stationary_distribution(result)
        else:
            raise NotImplementedError('nonreversible mle with fixed stationary distribution not implemented.')

    if return_statdist and isinstance(result, tuple):
        T, mu = result
    else:
        T = result

    # convert return type
    if sparse_computation and not sparse_input_type:
        T = T.toarray()
    elif not sparse_computation and sparse_input_type:
        T = csr_matrix(T)

    if return_statdist:
        return T, mu
    return T


def log_likelihood(C, T):
    r"""Log-likelihood of the count matrix given a transition matrix.

    Parameters
    ----------
    C : (M, M) ndarray or scipy.sparse matrix
        Count matrix
    T : (M, M) ndarray orscipy.sparse matrix
        Transition matrix

    Returns
    -------
    logL : float
        Log-likelihood of the count matrix

    Notes
    -----

    The likelihood of a set of observed transition counts
    :math:`C=(c_{ij})` for a given matrix of transition counts
    :math:`T=(t_{ij})` is given by

    .. math:: L(C|P)=\prod_{i=1}^{M} \left( \prod_{j=1}^{M} p_{ij}^{c_{ij}} \right)

    The log-likelihood is given by

    .. math:: l(C|P)=\sum_{i,j=1}^{M}c_{ij} \log p_{ij}.

    The likelihood describes the probability of making an observation
    :math:`C` for a given model :math:`P`.

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.estimation import log_likelihood

    >>> T = np.array([[0.9, 0.1, 0.0], [0.5, 0.0, 0.5], [0.0, 0.1, 0.9]])

    >>> C = np.array([[58, 7, 0], [6, 0, 4], [0, 3, 21]])
    >>> logL = log_likelihood(C, T)
    >>> logL # doctest: +ELLIPSIS
    -38.2808034725...

    >>> C = np.array([[58, 20, 0], [6, 0, 4], [0, 3, 21]])
    >>> logL = log_likelihood(C, T)
    >>> logL # doctest: +ELLIPSIS
    -68.2144096814...

    References
    ----------
    .. [1] Prinz, J H, H Wu, M Sarich, B Keller, M Senne, M Held, J D
        Chodera, C Schuette and F Noe. 2011. Markov models of
        molecular kinetics: Generation and validation. J Chem Phys
        134: 174105

    """
    if issparse(C) and issparse(T):
        return sparse.likelihood.log_likelihood(C, T)
    else:
        # use the dense likelihood calculator for all other cases
        # if a mix of dense/sparse C/T matrices is used, then both
        # will be converted to ndarrays.
        if not isinstance(C, np.ndarray):
            C = np.array(C)
        if not isinstance(T, np.ndarray):
            T = np.array(T)
        # computation is still efficient, because we only use terms
        # for nonzero elements of T
        nz = np.nonzero(T)
        return np.dot(C[nz], np.log(T[nz]))


def tmatrix_cov(C, k=None):
    r"""Covariance tensor for non-reversible transition matrix posterior.

    Parameters
    ----------
    C : (M, M) ndarray or scipy.sparse matrix
        Count matrix
    k : int (optional)
        Return only covariance matrix for entires in the k-th row of
        the transition matrix

    Returns
    -------
    cov : (M, M, M) ndarray
        Covariance tensor for transition matrix posterior

    Notes
    -----
    The posterior of non-reversible transition matrices is

    .. math:: \mathbb{P}(T|C) \propto \prod_{i=1}^{M} \left( \prod_{j=1}^{M} p_{ij}^{c_{ij}} \right)

    Each row in the transition matrix is distributed according to a
    Dirichlet distribution with parameters given by the observed
    transition counts :math:`c_{ij}`.

    The covariance tensor
    :math:`\text{cov}[p_{ij},p_{kl}]=\Sigma_{i,j,k,l}` is zero
    whenever :math:`i \neq k` so that only :math:`\Sigma_{i,j,i,l}` is
    returned.

    """
    if issparse(C):
        warnings.warn("Covariance matrix will be dense for sparse input")
        C = C.toarray()
    return dense.covariance.tmatrix_cov(C, row=k)


def error_perturbation(C, S):
    r"""Error perturbation for given sensitivity matrix.

    Parameters
    ----------
    C : (M, M) ndarray
        Count matrix
    S : (M, M) ndarray or (K, M, M) ndarray
        Sensitivity matrix (for scalar observable) or sensitivity
        tensor for vector observable

    Returns
    -------
    X : float or (K, K) ndarray
        error-perturbation (for scalar observables) or covariance matrix
        (for vector-valued observable)

    Notes
    -----

    **Scalar observable**

    The sensitivity matrix :math:`S=(s_{ij})` of a scalar observable
    :math:`f(T)` is defined as

    .. math:: S= \left(\left. \frac{\partial f(T)}{\partial t_{ij}} \right \rvert_{T_0} \right)

    evaluated at a suitable transition matrix :math:`T_0`.

    The sensitivity is the variance of the observable

    .. math:: \mathbb{V}(f)=\sum_{i,j,k,l} s_{ij} \text{cov}[t_{ij}, t_{kl}] s_{kl}

    **Vector valued observable**

    The sensitivity tensor :math:`S=(s_{ijk})` for a vector
    valued observable :math:`(f_1(T),\dots,f_K(T))` is defined as

    .. math:: S= \left( \left. \frac{\partial f_i(T)}{\partial t_{jk}} \right\rvert_{T_0} \right)

    evaluated at a suitable transition matrix :math:`T_0`.

    The sensitivity is the covariance matrix for the observable

    .. math:: \text{cov}[f_{\alpha}(T),f_{\beta}(T)] = \sum_{i,j,k,l} s_{\alpha i j}
                                                       \text{cov}[t_{ij}, t_{kl}] s_{\beta kl}

    """
    if issparse(C):
        warnings.warn("Error-perturbation will be dense for sparse input")
        C = C.toarray()
    return dense.covariance.error_perturbation(C, S)


def _showSparseConversionWarning():
    warnings.warn('Converting input to dense, since method is '
                  'currently only implemented for dense matrices.', UserWarning)


def sample_tmatrix(C, nsample=1, nsteps=None, reversible=False, mu=None, T0=None, return_statdist=False):
    r"""samples transition matrices from the posterior distribution

    Parameters
    ----------
    C : (M, M) ndarray or scipy.sparse matrix
        Count matrix
    nsample : int
        number of samples to be drawn
    nstep : int, default=None
        number of full Gibbs sampling sweeps internally done for each sample
        returned. This option is meant to ensure approximately uncorrelated
        samples for every call to sample(). If None, the number of steps will
        be automatically determined based on the other options and  the matrix
        size. nstep>1 will only be used for reversible sampling, because
        nonreversible sampling generates statistically independent transition
        matrices every step.
    reversible : bool
        If true sample from the ensemble of transition matrices
        restricted to those obeying a detailed balance condition,
        else draw from the whole ensemble of stochastic matrices.
    mu : array_like
        A fixed stationary distribution. Transition matrices with that stationary distribution will be sampled
    T0 : ndarray, shape=(n, n) or scipy.sparse matrix
        Starting point of the MC chain of the sampling algorithm.
        Has to obey the required constraints.
    return_statdist : bool, optional, default = False
        if true, will also return the stationary distribution.

    Returns
    -------
    P : ndarray(n,n) or array of ndarray(n,n)
        sampled transition matrix (or multiple matrices if nsample > 1)

    Notes
    -----
    The transition matrix sampler generates transition matrices from
    the posterior distribution. The posterior distribution is given as
    a product of Dirichlet distributions

    .. math:: \mathbb{P}(T|C) \propto \prod_{i=1}^{M} \left( \prod_{j=1}^{M} p_{ij}^{c_{ij}} \right)

    See also
    --------
    tmatrix_sampler

    """
    if issparse(C):
        _showSparseConversionWarning()
        C = C.toarray()

    sampler = tmatrix_sampler(C, reversible=reversible, mu=mu, T0=T0, nsteps=nsteps)
    return sampler.sample(nsamples=nsample, return_statdist=return_statdist)


def tmatrix_sampler(C, reversible=False, mu=None, T0=None, nsteps=None, prior='sparse'):
    r"""Generate transition matrix sampler object.

    Parameters
    ----------
    C : (M, M) ndarray or scipy.sparse matrix
        Count matrix
    reversible : bool
        If true sample from the ensemble of transition matrices
        restricted to those obeying a detailed balance condition,
        else draw from the whole ensemble of stochastic matrices.
    mu : array_like
        A fixed stationary distribution. Transition matrices with that
        stationary distribution will be sampled
    T0 : ndarray, shape=(n, n) or scipy.sparse matrix
        Starting point of the MC chain of the sampling algorithm.
        Has to obey the required constraints.
    nstep : int, default=None
        number of full Gibbs sampling sweeps per sample. This option is meant
        to ensure approximately uncorrelated samples for every call to
        sample(). If None, the number of steps will be automatically determined
        based on the other options and  the matrix size. nstep>1 will only be
        used for reversible sampling, because nonreversible sampling generates
        statistically independent transition matrices every step.

    Returns
    -------
    sampler : A :py:class:dense.tmatrix_sampler.TransitionMatrixSampler object that can be used to generate samples.

    Notes
    -----
    The transition matrix sampler generates transition matrices from
    the posterior distribution. The posterior distribution is given as
    a product of Dirichlet distributions

    .. math:: \mathbb{P}(T|C) \propto \prod_{i=1}^{M} \left( \prod_{j=1}^{M} p_{ij}^{c_{ij}} \right)

    The method can generate samples from the posterior under the following constraints

    **Reversible sampling**

    Using a MCMC sampler outlined in .. [1] it is ensured that samples
    from the posterior are reversible, i.e. there is a probability
    vector :math:`(\mu_i)` such that :math:`\mu_i t_{ij} = \mu_j
    t_{ji}` holds for all :math:`i,j`.

    **Reversible sampling with fixed stationary vector**

    Using a MCMC sampler outlined in .. [2] it is ensured that samples
    from the posterior fulfill detailed balance with respect to a given
    probability vector :math:`(\mu_i)`.

    References
    ----------
    .. [1] Noe, F. Probability distributions of molecular observables
        computed from Markov state models. J Chem Phys 128: 244103 (2008)

    .. [2] Trendelkamp-Schroer, B., H. Wu, F. Paul and F. Noe: Estimation and
        uncertainty of reversible Markov models. J. Chem. Phys. (submitted)

    """
    if issparse(C):
        _showSparseConversionWarning()
        C = C.toarray()

    from .dense.tmat_sampling.tmatrix_sampler import TransitionMatrixSampler
    sampler = TransitionMatrixSampler(C, reversible=reversible, mu=mu, P0=T0,
                                      n_steps=nsteps, prior=prior)
    return sampler


def rate_matrix(C, dt=1.0, method='KL', sparsity=None,
                t_agg=None, pi=None, tol=1.0E7, K0=None,
                maxiter=100000, on_error='raise'):
    r"""Estimate a reversible rate matrix from a count matrix.

    Parameters
    ----------
    C : (N,N) ndarray
        count matrix at a lag time dt
    dt : float, optional, default=1.0
        lag time that was used to estimate C
    method : str, one of {'KL', 'CVE', 'pseudo', 'truncated_log'}
        Method to use for estimation of the rate matrix.

        * 'pseudo' selects the pseudo-generator. A reversible transition
          matrix T is estimated and :math:`(T-Id)/d` is returned as the rate matrix.

        * 'truncated_log' selects the truncated logarithm :footcite:`davies2010embeddable`. A
          reversible transition matrix T is estimated and :math:`max(logm(T*T)/(2dt),0)`
          is returned as the rate matrix. logm is the matrix logarithm and
          the maximum is taken element-wise.

        * 'CVE' selects the algorithm of Crommelin and Vanden-Eijnden :footcite:`crommelin2009data`.
          It consists of minimizing the following objective function:

          .. math:: f(K)=\sum_{ij}\left(\sum_{kl} U_{ik}^{-1}K_{kl}U_{lj}-L_{ij}\right)^2 \left|\Lambda_{i}\Lambda_{j}\right|

          where :math:`\Lambda_i` are the eigenvalues of :math:`T` and :math:`U`
          is the matrix of its (right) eigenvectors; :math:`L_{ij}=\delta_{ij}\frac{1}{\tau}\log\left|\Lambda_i\right|`.
          :math:`T` is computed from C using the reversible maximum likelihood
          estimator.

        * 'KL' selects the algorihtm of Kalbfleisch and Lawless :footcite:`kalbfleisch1985analysis`.
          It consists of maximizing the following log-likelihood:

          .. math:: f(K)=\log L=\sum_{ij}C_{ij}\log(e^{K\Delta t})_{ij}

          where :math:`C_{ij}` are the transition counts at a lag-time :math:`\Delta t`.
          Here :math:`e` is the matrix exponential and the logarithm is taken
          element-wise.

    sparsity : (N,N) ndarray or None, optional, default=None
        If sparsity is None, a fully occupied rate matrix will be estimated.
        Alternatively, with the methods 'CVE' and 'KL' a ndarray of the
        same shape as C can be supplied. If sparsity[i,j]=0 and sparsity[j,i]=0
        the rate matrix elements :math:`K_{ij}` and :math:`K_{ji}` will be
        constrained to zero.
    t_agg : float, optional
        the aggregated simulation time;
        by default this is the total number of transition counts times
        the lag time (no sliding window counting). This value is used
        to compute the lower bound on the transition rate (that are not zero).
        If sparsity is None, this value is ignored.
    pi : (N) ndarray, optional
        the stationary vector of the desired rate matrix K.
        If no pi is given, the function takes the stationary vector
        of the MLE reversible T matrix that is computed from C.
    tol : float, optional, default = 1.0E7
        Tolerance of the quasi-Newton algorithm that is used to minimize
        the objective function. This is passed as the `factr` parameter to
        `scipy.optimize.fmin_l_bfgs_b`.
        Typical values for factr are: 1e12 for low accuracy; 1e7
        for moderate accuracy; 10.0 for extremely high accuracy.
    maxiter : int, optional, default = 100000
        Minimization of the objective function will do at most this number
        of steps.
    on_error : string, optional, default = 'raise'
        What to do then an error happend. When 'raise' is given, raise
        an exception. When 'warn' is given, produce a (Python) warning.

    Returns
    -------
    K : (N,N) ndarray
        the optimal rate matrix

    Notes
    -----
    In this implementation the algorithm of Crommelin and Vanden-Eijnden
    (CVE) is initialized with the pseudo-generator estimate. The
    algorithm of Kalbfleisch and Lawless (KL) is initialized using the
    CVE result.

    Example
    -------
    >>> import numpy as np
    >>> from deeptime.markov.tools.estimation import rate_matrix
    >>> C = np.array([[100,1],[50,50]])
    >>> rate_matrix(C)
    array([[-0.0138...,  0.0138...],
           [ 0.6993..., -0.6993...]])

    References
    ----------
    .. footbibliography::
    """

    from .dense.ratematrix import estimate_rate_matrix
    return estimate_rate_matrix(C, dt=dt, method=method, sparsity=sparsity,
                                t_agg=t_agg, pi=pi, tol=tol, K0=K0,
                                maxiter=maxiter, on_error=on_error)



import numpy as _np
import scipy as _sp
import scipy.sparse.linalg

__author__ = 'noe, clonker'


class ZeroRankError(_np.linalg.LinAlgError):
    """Input matrix has rank zero."""
    pass


def sort_eigs(evals, evecs, order='magnitude'):
    r""" Sorts the eigenvalues and eigenvectors by descending norm of the eigenvalues or lexicographically.

    Parameters
    ----------
    evals: (n, ) ndarray
        eigenvalues
    evecs: (k, n) ndarray
        eigenvectors in a column matrix
    order : str, default='magnitude'
        The order. Sorts by magnitude by default, can also be 'lexicographic' in which case it sorts lexicographically.

    Returns
    -------
    (evals, evecs) : (n, ) ndarray, (k, n) ndarray
        the sorted eigenvalues and eigenvectors
    """
    assert order in sort_eigs.supported_orders, f"Invalid order {order}, must be one of {sort_eigs.supported_orders}"
    sort_key = evals if order != 'magnitude' else _np.abs(evals)  # magnitude or plain depending on order
    order = _np.argsort(sort_key)[::-1]  # inverted argsort so that order is descending
    return evals[order], evecs[:, order]


sort_eigs.supported_orders = 'magnitude', 'lexicographic'  #: The order in which eigenvalues can be sorted.


def spd_truncated_svd(mat, dim=None, eps=0.):
    r""" Rank-reduced singular value decomposition of symmetric positive (semi-)definite matrix.
    The method yields for a matrix :math:`A\in\mathbb{R}^{n\times n}` singular values :math:`s\in\mathbb{R}^d`
    and singular vectors :math:`U\in\mathbb{R}^{n\times d}` so that `A \approx U\mathrm{diag}(x)U^\top`.

    All the negligible components are removed from the spectrum. In case of `dim` being specified it keeps
    at most the `dim` dominant components but may remove even more, depending on the input matrix.

    Eps influences the tolerance under which components are deemed negligible. In particular, if the product of
    `eps` and the largest singular value is larger than a component of the spectrum, it is removed.

    Parameters
    ----------
    mat : (n, n) ndarray
        Input matrix.
    dim : int, optional, default=None
        The dimension.
    eps : float, optional, default = 0
        Tolerance.

    Returns
    -------
    s : (k, ) ndarray
        Leading singular values.
    U : (n, k) ndarray
        Leading singular vectors.
    """
    if dim is None:
        dim = _np.inf
    dim = min(dim, mat.shape[0])
    S, U = scipy.linalg.schur(mat)
    s = _np.diag(S)
    max_sv = _np.abs(s).max()
    tol = mat.shape[0] * _np.spacing(max_sv)
    mi = _np.min(s)
    if mi < 0 and -mi > tol:
        tol = -mi
    tol = _np.maximum(tol, max_sv * eps)
    dim = min(dim, _np.count_nonzero(s > tol))
    idx = s.argsort()[::-1][:dim]
    return s[idx], U[:, idx]


def spd_eig(W, epsilon=1e-10, method='QR', canonical_signs=False, check_sym: bool = False):
    """ Rank-reduced eigenvalue decomposition of symmetric positive definite matrix.

    Removes all negligible eigenvalues

    Parameters
    ----------
    W : ndarray((n, n), dtype=float)
        Symmetric positive-definite (spd) matrix.
    epsilon : float
        Truncation parameter. Eigenvalues with norms smaller than this cutoff will
        be removed.
    method : str
        Method to perform the decomposition of :math:`W` before inverting. Options are:

        * 'QR': QR-based robust eigenvalue decomposition of W
        * 'schur': Schur decomposition of W
    canonical_signs : bool, default = False
        Fix signs in V, such that the largest element in every column of V is positive.
    check_sym : bool, default = False
        Check whether the input matrix is (almost) symmetric.

    Returns
    -------
    s : ndarray(k)
        k non-negligible eigenvalues, sorted by descending norms

    V : ndarray(n, k)
        k leading eigenvectors
    """
    # check input
    if check_sym and not _np.allclose(W.T, W):
        raise ValueError('W is not a symmetric matrix')

    if method == 'QR':
        from .eig_qr import eig_qr
        s, V = eig_qr(W)
    # compute the Eigenvalues of C0 using Schur factorization
    elif method == 'schur':
        from scipy.linalg import schur
        S, V = schur(W)
        s = _np.diag(S)
    else:
        raise ValueError(f'method {method} not implemented, available are {spd_eig.methods}')

    s, V = sort_eigs(s, V)  # sort them

    # determine the cutoff. We know that C0 is an spd matrix,
    # so we select the truncation threshold such that everything that is negative vanishes
    evmin = _np.min(s)
    if evmin < 0:
        epsilon = max(epsilon, -evmin + 1e-16)

    # determine effective rank m and perform low-rank approximations.
    evnorms = _np.abs(s)
    n = _np.shape(evnorms)[0]
    m = n - _np.searchsorted(evnorms[::-1], epsilon)
    if m == 0:
        raise ZeroRankError(
            'All eigenvalues are smaller than %g, rank reduction would discard all dimensions.' % epsilon)
    Vm = V[:, 0:m]
    sm = s[0:m]

    if canonical_signs:
        # enforce canonical eigenvector signs
        for j in range(m):
            jj = _np.argmax(_np.abs(Vm[:, j]))
            Vm[:, j] *= _np.sign(Vm[jj, j])

    return sm, Vm


spd_eig.methods = ('QR', 'schur')


def spd_inv(W, epsilon=1e-10, method='QR'):
    """
    Compute matrix inverse of symmetric positive-definite matrix :math:`W`.

    by first reducing W to a low-rank approximation that is truly spd
    (Moore-Penrose inverse).

    Parameters
    ----------
    W : ndarray((m,m), dtype=float)
        Symmetric positive-definite (spd) matrix.
    epsilon : float
        Truncation parameter. Eigenvalues with norms smaller than this cutoff will
        be removed.
    method : str
        Method to perform the decomposition of :math:`W` before inverting. Options are:

        * 'QR': QR-based robust eigenvalue decomposition of W
        * 'schur': Schur decomposition of W

    Returns
    -------
    L : ndarray((n, r))
        the Moore-Penrose inverse of the symmetric positive-definite matrix :math:`W`

    """
    if _np.shape(W)[0] == 1:
        if W[0, 0] < epsilon:
            raise ZeroRankError(
                'All eigenvalues are smaller than %g, rank reduction would discard all dimensions.' % epsilon)
        Winv = 1. / W[0, 0]
    else:
        sm, Vm = spd_eig(W, epsilon=epsilon, method=method)
        Winv = _np.dot(Vm, _np.diag(1.0 / sm)).dot(Vm.T)

    # return split
    return Winv


def spd_inv_sqrt(W, epsilon=1e-10, method='QR', return_rank=False):
    """
    Computes :math:`W^{-1/2}` of symmetric positive-definite matrix :math:`W`.

    by first reducing W to a low-rank approximation that is truly spd.

    Parameters
    ----------
    W : ndarray((m,m), dtype=float)
        Symmetric positive-definite (spd) matrix.
    epsilon : float, optional, default=1e-10
        Truncation parameter. Eigenvalues with norms smaller than this cutoff will be removed.
    method : str, optional, default='QR'
        Method to perform the decomposition of :math:`W` before inverting. Options are:

        * 'QR': QR-based robust eigenvalue decomposition of W
        * 'schur': Schur decomposition of W

    return_rank : bool, optional, default=False
        Whether to return the rank of W.

    Returns
    -------
    L : ndarray((n, r))
        :math:`W^{-1/2}` after reduction of W to a low-rank spd matrix

    """
    if _np.shape(W)[0] == 1:
        if W[0, 0] < epsilon:
            raise ZeroRankError(
                'All eigenvalues are smaller than %g, rank reduction would discard all dimensions.' % epsilon)
        Winv = 1. / _np.sqrt(W[0, 0])
        sm = _np.ones(1)
    else:
        sm, Vm = spd_eig(W, epsilon=epsilon, method=method)
        Winv = _np.dot(Vm, _np.diag(1.0 / _np.sqrt(sm))).dot(Vm.T)

    # return split
    if return_rank:
        return Winv, sm.shape[0]
    else:
        return Winv


def spd_inv_split(W, epsilon=1e-10, method='QR', canonical_signs=False):
    """
    Compute :math:`W^{-1} = L L^T` of the symmetric positive-definite matrix :math:`W`.

    by first reducing W to a low-rank approximation that is truly spd.

    Parameters
    ----------
    W : ndarray((m,m), dtype=float)
        Symmetric positive-definite (spd) matrix.
    epsilon : float
        Truncation parameter. Eigenvalues with norms smaller than this cutoff will
        be removed.
    method : str
        Method to perform the decomposition of :math:`W` before inverting. Options are:

        * 'QR': QR-based robust eigenvalue decomposition of W
        * 'schur': Schur decomposition of W

    canonical_signs : boolean, default = False
        Fix signs in L, s. t. the largest element of in every row of L is positive.


    Returns
    -------
    L : ndarray((n, r))
        Matrix :math:`L` from the decomposition :math:`W^{-1} = L L^T`.

    """
    if _np.shape(W)[0] == 1:
        if W[0, 0] < epsilon:
            raise ZeroRankError(
                'All eigenvalues are smaller than %g, rank reduction would discard all dimensions.' % epsilon)
        L = 1. / _np.sqrt(W[0, 0])
    else:
        if epsilon == 0.:
            sm, Vm = _np.linalg.eigh(.5 * (W + W.T))
            sm, Vm = sort_eigs(sm, Vm)
            # threshold eigenvalues to be >= 0 and sqrt of the eigenvalues to be >= 1e-16 so that no
            # division by zero can occur
            L = _np.dot(Vm, _np.diag(1.0 / _np.maximum(_np.sqrt(_np.maximum(sm, 1e-12)), 1e-12)))
        else:
            sm, Vm = spd_eig(W, epsilon=epsilon, method=method, canonical_signs=canonical_signs)
            L = _np.dot(Vm, _np.diag(1.0 / _np.sqrt(sm)))

    # return split
    return L


def eigs(matrix: _np.ndarray, n_eigs=None, which='LM'):
    r"""Computes the eigenvalues and eigenvectors of `matrix`. Optionally the number of eigenvalue-eigenvector pairs
    can be provided, in which case they are computed using the Lanczos algorithm.

    Parameters
    ----------
    matrix : (n, n) ndarray
        The matrix to compute the eigenvalues and eigenvectors of.
    n_eigs : int, optional, default=None
        The number of eigenpairs. Can be None, in which case all eigenvalues are computed.
    which : str, default='LM'
        If `n_eigs` is provided, determines which eigenvalues are returned. Default is 'Largest Magnitude.
        See the `scipy <https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.eigs.html>`__
        documentation for all options and effects.

    Returns
    -------
    (eigenvalues, eigenvectors) : ((n, ) ndarray, (k, n) ndarray)
        The computed eigenvalues and eigenvectors.
    """
    n = matrix.shape[0]
    if n_eigs is not None and n_eigs < n:
        return _sp.sparse.linalg.eigs(matrix, n_eigs, which=which)
    else:
        return _sp.linalg.eig(matrix)


def eig_corr(C0, Ct, epsilon=1e-10, method='QR', canonical_signs=False, return_rank=False):
    r""" Solve generalized eigenvalue problem with correlation matrices C0 and Ct

    Numerically robust solution of a generalized Hermitian (symmetric) eigenvalue
    problem of the form

    .. math::
        \mathbf{C}_t \mathbf{r}_i = \mathbf{C}_0 \mathbf{r}_i l_i

    Computes :math:`m` dominant eigenvalues :math:`l_i` and eigenvectors
    :math:`\mathbf{r}_i`, where :math:`m` is the numerical rank of the problem.
    This is done by first conducting a Schur decomposition of the symmetric
    positive matrix :math:`\mathbf{C}_0`, then truncating its spectrum to
    retain only eigenvalues that are numerically greater than zero, then using
    this decomposition to define an ordinary eigenvalue Problem for
    :math:`\mathbf{C}_t` of size :math:`m`, and then solving this eigenvalue
    problem.

    Parameters
    ----------
    C0 : ndarray (n,n)
        time-instantaneous correlation matrix. Must be symmetric positive definite
    Ct : ndarray (n,n)
        time-lagged correlation matrix. Must be symmetric
    epsilon : float
        eigenvalue norm cutoff. Eigenvalues of C0 with norms <= epsilon will be
        cut off. The remaining number of Eigenvalues define the size of
        the output.
    method : str
        Method to perform the decomposition of :math:`W` before inverting. Options are:

        * 'QR': QR-based robust eigenvalue decomposition of W
        * 'schur': Schur decomposition of W
    canonical_signs : bool
        If True, re-scale each eigenvector such that its entry with maximal absolute value
        is positive.
    return_rank : bool, default=False
        If True, return the rank of generalized eigenvalue problem.

    Returns
    -------
    l : ndarray (m)
        The first m generalized eigenvalues, sorted by descending norm
    R : ndarray (n,m)
        The first m generalized eigenvectors, as a column matrix.
    rank: int
        Rank of :math:`C0^{-0.5}`, if return_rank is True.
    """
    L = spd_inv_split(C0, epsilon=epsilon, method=method, canonical_signs=True)
    Ct_trans = _np.dot(_np.dot(L.T, Ct), L)

    # solve the symmetric eigenvalue problem in the new basis
    if _np.allclose(Ct.T, Ct):
        from scipy.linalg import eigh
        l, R_trans = eigh(Ct_trans)
    else:
        from scipy.linalg import eig
        l, R_trans = eig(Ct_trans)

    # sort eigenpairs
    l, R_trans = sort_eigs(l, R_trans)

    # transform the eigenvectors back to the old basis
    R = _np.dot(L, R_trans)

    # Change signs of eigenvectors:
    if canonical_signs:
        for j in range(R.shape[1]):
            imax = _np.argmax(_np.abs(R[:, j]))
            R[:, j] *= _np.sign(R[imax, j])

    if return_rank:
        return l, R, L.shape[1] if L.ndim == 2 else 1

    # return result
    return l, R



import collections
from copy import deepcopy
from typing import Optional, Union, List

import numpy as np

from deeptime.markov.tools.analysis import is_connected
from deeptime.markov.tools.estimation import sample_tmatrix, transition_matrix

from deeptime.base import Estimator
from deeptime.markov._base import BayesianPosterior, MembershipsChapmanKolmogorovValidator
from deeptime.markov._transition_matrix import stationary_distribution
from deeptime.markov.hmm import HiddenMarkovModel
from ._output_model import DiscreteOutputModel
from ._util import observations_in_state, sample_hidden_state_trajectory
from deeptime.markov.msm import MarkovStateModel
from deeptime.markov import TransitionCountModel, compute_dtrajs_effective, number_of_states
from deeptime.markov.hmm._hmm_bindings import util as _bd_util
from deeptime.util.types import ensure_dtraj_list

__author__ = 'noe, clonker'

__all__ = [
    'BayesianHMMPosterior',
    'BayesianHMM',
]

from ...util.platform import handle_progress_bar


class BayesianHMMPosterior(BayesianPosterior):
    r""" Bayesian Hidden Markov model with samples of posterior and prior.

    Parameters
    ----------
    prior : HiddenMarkovModel, optional, default=None
        The prior.
    samples : list of HiddenMarkovModel, optional, default=()
        Sampled models.

    See Also
    --------
    BayesianHMM
    """

    def __init__(self,
                 prior: Optional[HiddenMarkovModel] = None,
                 samples: Optional[List[HiddenMarkovModel]] = ()):
        super(BayesianHMMPosterior, self).__init__(prior=prior, samples=samples)

    def submodel_largest(self, directed=True, connectivity_threshold='1/n', observe_nonempty=True, dtrajs=None):
        r""" Creates a submodel from the largest connected set.

        Parameters
        ----------
        directed : bool, optional, default=True
            Whether the connectivity graph on the count matrix is interpreted as directed.
        connectivity_threshold : float or '1/n', optional, default='1/n'.
            Connectivity threshold. counts that are below the specified value are disregarded when finding connected
            sets. In case of '1/n', the threshold gets resolved to :math:`1 / n\_states\_full`.
        observe_nonempty : bool, optional, default=True
            Whether to restrict to observable states which are observed in provided dtrajs. If True, dtrajs must not
            be None.
        dtrajs : array_like or list of array_like, optional, default=None
            Time series on which is evaluated whether observable states in the model were actually observed.

        Returns
        -------
        submodel : BayesianHMMPosterior
            The submodel.
        """
        dtrajs = ensure_dtraj_list(dtrajs)
        states = self.prior.states_largest(directed=directed, connectivity_threshold=connectivity_threshold)
        obs = self.prior.nonempty_obs(dtrajs) if observe_nonempty else None
        return self.submodel(states=states, obs=obs)

    def submodel_populous(self, directed=True, connectivity_threshold='1/n', observe_nonempty=True, dtrajs=None):
        r""" Creates a submodel from the most populated connected set.

        Parameters
        ----------
        directed : bool, optional, default=True
            Whether the connectivity graph on the count matrix is interpreted as directed.
        connectivity_threshold : float or '1/n', optional, default='1/n'.
            Connectivity threshold. counts that are below the specified value are disregarded when finding connected
            sets. In case of '1/n', the threshold gets resolved to :math:`1 / n\_states\_full`.
        observe_nonempty : bool, optional, default=True
            Whether to restrict to observable states which are observed in provided dtrajs. If True, dtrajs must not
            be None.
        dtrajs : array_like or list of array_like, optional, default=None
            Time series on which is evaluated whether observable states in the model were actually observed and
            which states were the most populated.

        Returns
        -------
        submodel : BayesianHMMPosterior
            The submodel.
        """
        dtrajs = ensure_dtraj_list(dtrajs)
        states = self.prior.states_populous(strong=directed, connectivity_threshold=connectivity_threshold)
        obs = self.prior.nonempty_obs(dtrajs) if observe_nonempty else None
        return self.submodel(states=states, obs=obs)

    @property
    def hidden_state_trajectories_samples(self):
        r""" Hidden state trajectories of sampled HMMs. Available if the estimator was configured to save them,
        see :attr:`BayesianHMM.store_hidden`.
        """
        return [s.hidden_state_trajectories for s in self.samples]

    @property
    def samples(self) -> Optional[List[HiddenMarkovModel]]:
        r""" The sampled models. """
        return self._samples

    @samples.setter
    def samples(self, value: Optional[List[HiddenMarkovModel]]):
        self._samples = value

    @property
    def prior(self) -> HiddenMarkovModel:
        r""" The prior model. """
        return self._prior

    @prior.setter
    def prior(self, value: HiddenMarkovModel):
        self._prior = value

    def submodel(self, states=None, obs=None):
        r""" Creates a submodel from this model restricted to a selection of observable and hidden states.

        Parameters
        ----------
        states : ndarray or None, optional, default=None
            The hidden states to restrict to. If None there is no restriction.
        obs : ndarray or None, optional, default=None
            The observable states to restrict to. If None there is no restriction. Only makes sense with
            :class:`DiscreteOutputModel`.

        Returns
        -------
        submodel : BayesianHMMPosterior
            The submodel.
        """
        # restrict prior
        sub_model = self.prior.submodel(states=states, obs=obs)
        # restrict reduce samples
        subsamples = [sample.submodel(states=states, obs=obs)
                      for sample in self]
        return BayesianHMMPosterior(sub_model, subsamples)


class BayesianHMM(Estimator):
    r""" Estimator for a Bayesian Hidden Markov state model.

    The theory and estimation procedure are described
    in :footcite:`noe2013projected`, :footcite:`chodera2011bayesian`.

    Parameters
    ----------
    initial_hmm : :class:`HMM <deeptime.markov.hmm.HiddenMarkovModel>`
       Single-point estimate of HMM object around which errors will be evaluated.
       There is a static method available that can be used to generate a default prior, see :meth:`default`.
    n_samples : int, optional, default=100
       Number of sampled models.
    stride : str or int, default='effective'
       stride between two lagged trajectories extracted from the input
       trajectories. Given trajectory s[t], stride and lag will result
       in trajectories

           :code:`s[0], s[tau], s[2 tau], ...`

           :code:`s[stride], s[stride + tau], s[stride + 2 tau], ...`

       Setting stride = 1 will result in using all data (useful for
       maximum likelihood estimator), while a Bayesian estimator requires
       a longer stride in order to have statistically uncorrelated
       trajectories. Setting stride = 'effective' uses the largest
       neglected timescale as an estimate for the correlation time and
       sets the stride accordingly.
    initial_distribution_prior : None, str, float or ndarray(n)
       Prior for the initial distribution of the HMM. Will only be active
       if stationary=False (stationary=True means that p0 is identical to
       the stationary distribution of the transition matrix).
       Currently implements different versions of the Dirichlet prior that
       is conjugate to the Dirichlet distribution of p0. p0 is sampled from:

       .. math::
           p0 \sim \prod_i (p0)_i^{a_i + n_i - 1}

       where :math:`n_i` are the number of times a hidden trajectory was in
       state :math:`i` at time step 0 and :math:`a_i` is the prior count.
       Following options are available:

       * 'mixed' (default),  :math:`a_i = p_{0,init}`, where :math:`p_{0,init}`
         is the initial distribution of initial_model.
       *  ndarray(n) or float,
          the given array will be used as A.
       *  'uniform',  :math:`a_i = 1`
       *   None,  :math:`a_i = 0`. This option ensures coincidence between
           sample mean an MLE. Will sooner or later lead to sampling problems,
           because as soon as zero trajectories are drawn from a given state,
           the sampler cannot recover and that state will never serve as a starting
           state subsequently. Only recommended in the large data regime and
           when the probability to sample zero trajectories from any state
           is negligible.
    transition_matrix_prior : str or ndarray(n, n)
       Prior for the HMM transition matrix.
       Currently implements Dirichlet priors if reversible=False and reversible
       transition matrix priors as described in :footcite:`trendelkamp2015estimation` if reversible=True. For the
       nonreversible case the posterior of transition matrix :math:`P` is:

       .. math::
           P \sim \prod_{i,j} p_{ij}^{b_{ij} + c_{ij} - 1}

       where :math:`c_{ij}` are the number of transitions found for hidden
       trajectories and :math:`b_{ij}` are prior counts.

       * 'mixed' (default),  :math:`b_{ij} = p_{ij,\mathrm{init}}`, where :math:`p_{ij,\mathrm{init}}`
         is the transition matrix of initial_model. That means one prior
         count will be used per row.
       * ndarray(n, n) or broadcastable,
         the given array will be used as B.
       * 'uniform',  :math:`b_{ij} = 1`
       * None,  :math:`b_ij = 0`. This option ensures coincidence between
         sample mean an MLE. Will sooner or later lead to sampling problems,
         because as soon as a transition :math:`ij` will not occur in a
         sample, the sampler cannot recover and that transition will never
         be sampled again. This option is not recommended unless you have
         a small HMM and a lot of data.
    store_hidden : bool, optional, default=False
       Store hidden trajectories in sampled HMMs, see
       :attr:`BayesianHMMPosterior.hidden_state_trajectories_samples`.
    reversible : bool, optional, default=True
       If True, a prior that enforces reversible transition matrices (detailed balance) is used;
       otherwise, a standard  non-reversible prior is used.
    stationary : bool, optional, default=False
       If True, the stationary distribution of the transition matrix will be used as initial distribution.
       Only use True if you are confident that the observation trajectories are started from a global
       equilibrium. If False, the initial distribution will be estimated as usual from the first step
       of the hidden trajectories.

    References
    ----------
    .. footbibliography::
    """

    _SampleStorage = collections.namedtuple(
        '_SampleStorage', ['transition_matrix', 'output_model', 'stationary_distribution', 'initial_distribution',
                           'counts', 'hidden_trajs']
    )

    def __init__(self, initial_hmm: HiddenMarkovModel,
                 n_samples: int = 100,
                 n_transition_matrix_sampling_steps: int = 1000,
                 stride: Union[str, int] = 'effective',
                 initial_distribution_prior: Optional[Union[str, float, np.ndarray]] = 'mixed',
                 transition_matrix_prior: Optional[Union[str, np.ndarray]] = 'mixed',
                 store_hidden: bool = False,
                 reversible: bool = True,
                 stationary: bool = False):
        super().__init__()
        self.initial_hmm = initial_hmm
        self.n_samples = n_samples
        self.n_transition_matrix_sampling_steps = n_transition_matrix_sampling_steps
        self.stride = stride
        self.initial_distribution_prior = initial_distribution_prior
        self.transition_matrix_prior = transition_matrix_prior
        self.store_hidden = store_hidden
        self.reversible = reversible
        self.stationary = stationary

    @staticmethod
    def default(dtrajs, n_hidden_states: int, lagtime: int, n_samples: int = 100,
                stride: Union[str, int] = 'effective',
                initial_distribution_prior: Optional[Union[str, float, np.ndarray]] = 'mixed',
                transition_matrix_prior: Optional[Union[str, np.ndarray]] = 'mixed',
                separate: Optional[Union[int, List[int]]] = None,
                store_hidden: bool = False,
                reversible: bool = True,
                stationary: bool = False,
                prior_submodel: bool = True):
        """ Computes a default prior for a BHMM and uses that for error estimation.
        For a more detailed description of the arguments please
        refer to :class:`HMM <deeptime.markov.hmm.HiddenMarkovModel>` or
        :meth:`__init__`.

        Returns
        -------
        estimator : BayesianHMM
            Estimator that is initialized with a default prior model.
        """
        from deeptime.markov.hmm import init, MaximumLikelihoodHMM
        dtrajs = ensure_dtraj_list(dtrajs)
        init_hmm = init.discrete.metastable_from_data(dtrajs, n_hidden_states=n_hidden_states, lagtime=lagtime,
                                                      stride=stride, reversible=reversible, stationary=stationary,
                                                      separate_symbols=separate)
        hmm = MaximumLikelihoodHMM(init_hmm, stride=stride, lagtime=lagtime, reversible=reversible,
                                   stationary=stationary, accuracy=1e-2).fit(dtrajs).fetch_model()
        if prior_submodel:
            hmm = hmm.submodel_largest(connectivity_threshold=0, observe_nonempty=False, dtrajs=dtrajs)
        estimator = BayesianHMM(hmm, n_samples=n_samples, stride=stride,
                                initial_distribution_prior=initial_distribution_prior,
                                transition_matrix_prior=transition_matrix_prior, store_hidden=store_hidden,
                                reversible=reversible, stationary=stationary)
        return estimator

    @property
    def stationary(self):
        r""" If True, the stationary distribution of the transition matrix will be used as initial distribution.
        Only use True if you are confident that the observation trajectories are started from a global
        equilibrium. If False, the initial distribution will be estimated as usual from the first step
        of the hidden trajectories.
        """
        return self._stationary

    @stationary.setter
    def stationary(self, value):
        self._stationary = value

    @property
    def reversible(self):
        r""" If True, a prior that enforces reversible transition matrices (detailed balance) is used;
        otherwise, a standard  non-reversible prior is used.
        """
        return self._reversible

    @reversible.setter
    def reversible(self, value):
        self._reversible = value

    @property
    def store_hidden(self):
        r"""  Store hidden trajectories in sampled HMMs, see
        :attr:`BayesianHMMPosterior.hidden_state_trajectories_samples`.
        """
        return self._store_hidden

    @store_hidden.setter
    def store_hidden(self, value):
        self._store_hidden = value

    @property
    def transition_matrix_prior(self):
        r""" Prior for the transition matrix. For a more detailed description refer to :meth:`__init__`. """
        return self._transition_matrix_prior

    @transition_matrix_prior.setter
    def transition_matrix_prior(self, value):
        self._transition_matrix_prior = value

    @property
    def initial_distribution_prior(self) -> Optional[Union[str, float, np.ndarray]]:
        r""" Prior for the initial distribution. For a more detailed description refer to :meth:`__init__`. """
        return self._initial_distribution_prior

    @initial_distribution_prior.setter
    def initial_distribution_prior(self, value: Optional[Union[str, float, np.ndarray]]):
        self._initial_distribution_prior = value

    @property
    def initial_hmm(self) -> HiddenMarkovModel:
        r""" The prior HMM. An estimator with a default prior HMM can be generated using the static :meth:`default`
        method.
        """
        return self._initial_hmm

    @initial_hmm.setter
    def initial_hmm(self, value: HiddenMarkovModel):
        if not isinstance(value, HiddenMarkovModel):
            raise ValueError(f"Initial hmm must be of type HiddenMarkovModel, but was {value.__class__.__name__}.")
        self._initial_hmm = value

    @property
    def n_samples(self) -> int:
        r""" Number of sampled models. """
        return self._n_samples

    @n_samples.setter
    def n_samples(self, value: int):
        self._n_samples = value

    def fetch_model(self) -> BayesianHMMPosterior:
        r""" Yields the current model or None if :meth:`fit` was not yet called.

        Returns
        -------
        posterior : BayesianHMMPosterior
            The model.
        """
        return self._model

    @property
    def _initial_distribution_prior_np(self) -> np.ndarray:
        r"""
        Internal method that evaluates the prior to its ndarray realization.
        """
        if self.initial_distribution_prior is None or self.initial_distribution_prior == 'sparse':
            prior = np.zeros(self.initial_hmm.n_hidden_states)
        elif isinstance(self.initial_distribution_prior, np.ndarray):
            if self.initial_distribution_prior.ndim == 1 \
                    and len(self.initial_distribution_prior) == self.initial_hmm.n_hidden_states:
                prior = np.asarray(self.initial_distribution_prior)
            else:
                raise ValueError(f"If the initial distribution prior is given as a np array, it must be 1-dimensional "
                                 f"and be as long as there are hidden states. "
                                 f"(ndim={self.initial_distribution_prior.ndim}, "
                                 f"len={len(self.initial_distribution_prior)})")
        elif self.initial_distribution_prior == 'mixed':
            prior = self.initial_hmm.initial_distribution
        elif self.initial_distribution_prior == 'uniform':
            prior = np.ones(self.initial_hmm.n_hidden_states)
        else:
            raise ValueError(f'Initial distribution prior mode undefined: {self.initial_distribution_prior}')
        return prior

    @property
    def _transition_matrix_prior_np(self) -> np.ndarray:
        r"""
        Internal method that evaluates the prior to its ndarray realization.
        """
        n_states = self.initial_hmm.n_hidden_states
        if self.transition_matrix_prior is None or self.transition_matrix_prior == 'sparse':
            prior = np.zeros((n_states, n_states))
        elif isinstance(self.transition_matrix_prior, np.ndarray):
            if np.array_equal(self.transition_matrix_prior.shape, (n_states, n_states)):
                prior = np.asarray(self.transition_matrix_prior)
            else:
                raise ValueError(f"If the initial distribution prior is given as a np array, it must be 2-dimensional "
                                 f"and a (n_hidden_states x n_hidden_states) matrix. "
                                 f"(ndim={self.transition_matrix_prior.ndim}, "
                                 f"shape={self.transition_matrix_prior.shape})")
        elif self.transition_matrix_prior == 'mixed':
            prior = np.copy(self.initial_hmm.transition_model.transition_matrix)
        elif self.transition_matrix_prior == 'uniform':
            prior = np.ones((n_states, n_states))
        else:
            raise ValueError(f'Initial distribution prior mode undefined: {self.transition_matrix_prior}')
        return prior

    def _update(self, model: _SampleStorage, observations, temp_alpha,
                transition_matrix_prior, initial_distribution_prior):
        """Update the current model using one round of Gibbs sampling."""
        self._update_hidden_state_trajectories(model, observations, temp_alpha)
        self._update_emission_probabilities(model, observations)
        self._update_transition_matrix(model, transition_matrix_prior, initial_distribution_prior,
                                       reversible=self.reversible, stationary=self.stationary,
                                       n_sampling_steps=self.n_transition_matrix_sampling_steps)

    @staticmethod
    def _update_hidden_state_trajectories(model: _SampleStorage, observations, temp_alpha):
        """Sample a new set of state trajectories from the conditional distribution P(S | T, E, O)"""
        model.hidden_trajs.clear()
        for obs in observations:
            s_t = sample_hidden_state_trajectory(model.transition_matrix, model.output_model,
                                                 model.initial_distribution, obs, temp_alpha)
            model.hidden_trajs.append(s_t)

    @staticmethod
    def _update_emission_probabilities(model: _SampleStorage, observations):
        """Sample a new set of emission probabilites from the conditional distribution P(E | S, O) """
        observations_by_state = [observations_in_state(model.hidden_trajs, observations, state)
                                 for state in range(model.transition_matrix.shape[0])]
        model.output_model.sample(observations_by_state)

    @staticmethod
    def _update_transition_matrix(model: _SampleStorage, transition_matrix_prior,
                                  initial_distribution_prior, reversible: bool = True, stationary: bool = False,
                                  n_sampling_steps: int = 1000):
        """ Updates the hidden-state transition matrix and the initial distribution """
        C = _bd_util.count_matrix(model.hidden_trajs, 1, model.transition_matrix.shape[0])
        model.counts[...] = C
        C = C + transition_matrix_prior

        # check if we work with these options
        if reversible and not is_connected(C, directed=True):
            raise NotImplementedError('Encountered disconnected count matrix with sampling option reversible:\n '
                                      f'{C}\nUse prior to ensure connectivity or use reversible=False.')
        # ensure consistent sparsity pattern (P0 might have additional zeros because of underflows)
        # TODO: these steps work around a bug in msmtools. Should be fixed there
        P0 = transition_matrix(C, reversible=reversible, maxiter=10000, warn_not_converged=False)
        zeros = np.where(P0 + P0.T == 0)
        C[zeros] = 0
        # run sampler
        Tij = sample_tmatrix(C, nsample=1, nsteps=n_sampling_steps, reversible=reversible)

        # INITIAL DISTRIBUTION
        if stationary:  # p0 is consistent with P
            p0 = stationary_distribution(Tij, C=C)
        else:
            n0 = BayesianHMM._count_init(model.hidden_trajs, model.transition_matrix.shape[0])
            first_timestep_counts_with_prior = n0 + initial_distribution_prior
            positive = first_timestep_counts_with_prior > 0
            p0 = np.zeros(n0.shape, dtype=model.transition_matrix.dtype)
            p0[positive] = np.random.dirichlet(first_timestep_counts_with_prior[positive])  # sample p0 from posterior

        # update HMM with new sample
        model.transition_matrix[...] = Tij
        model.stationary_distribution[...] = p0

    @staticmethod
    def _count_init(hidden_state_trajectories, n_hidden_states):
        """Compute the counts at the first time step

        Returns
        -------
        n : ndarray(nstates)
            n[i] is the number of trajectories starting in state i

        """
        n = [traj[0] for traj in hidden_state_trajectories]
        return np.bincount(n, minlength=n_hidden_states)

    def fit(self, data, n_burn_in: int = 0, n_thin: int = 1, progress=None, **kwargs):
        r""" Sample from the posterior.

        Parameters
        ----------
        data : array_like or list of array_like
            Input time series data.
        n_burn_in : int, optional, default=0
            The number of samples to discard to burn-in, following which :attr:`n_samples` samples will be generated.
        n_thin : int, optional, default=1
            The number of Gibbs sampling updates used to generate each returned sample.
        progress : iterable, optional, default=None
            Optional progressbar. Tested for tqdm.
        **kwargs
            Ignored kwargs for scikit-learn compatibility.

        Returns
        -------
        self : BayesianHMM
            Reference to self.
        """
        progress = handle_progress_bar(progress)
        dtrajs = ensure_dtraj_list(data)

        # fetch priors
        tmat = self.initial_hmm.transition_model.transition_matrix
        transition_matrix_prior = self._transition_matrix_prior_np

        initial_distribution_prior = self._initial_distribution_prior_np

        model = BayesianHMMPosterior()
        # update HMM Model
        model.prior = self.initial_hmm.copy()

        prior = model.prior

        # check if we are strongly connected in the reversible case (plus prior)
        if self.reversible and not is_connected(tmat + transition_matrix_prior, directed=True):
            raise NotImplementedError('Trying to sample disconnected HMM with option reversible:\n '
                                      f'{tmat}\n Use prior to connect, select connected subset, '
                                      f'or use reversible=False.')

        # EVALUATE STRIDE
        dtrajs_lagged_strided = compute_dtrajs_effective(
            dtrajs, lagtime=prior.lagtime, n_states=prior.n_hidden_states, stride=self.stride
        )
        # if stride is different to init_hmm, check if microstates in lagged-strided trajs are compatible
        if self.stride != self.initial_hmm.stride:
            symbols = np.unique(np.concatenate(dtrajs_lagged_strided))
            if not len(np.intersect1d(self.initial_hmm.observation_symbols, symbols)) == len(symbols):
                raise ValueError('Choice of stride has excluded a different set of microstates than in '
                                 'init_hmm. Set of observed microstates in time-lagged strided trajectories '
                                 'must match to the one used for init_hmm estimation.')

        # here we blow up the output matrix (if needed) to the FULL state space because we want to use dtrajs in the
        # Bayesian HMM sampler. This is just an initialization.
        n_states_full = number_of_states(dtrajs_lagged_strided)

        if prior.n_observation_states < n_states_full:
            eps = 0.01 / n_states_full  # default output probability, in order to avoid zero columns
            # full state space output matrix. make sure there are no zero columns
            full_obs_probabilities = eps * np.ones((prior.n_hidden_states, n_states_full), dtype=np.float64)
            # fill active states
            full_obs_probabilities[:, prior.observation_symbols] = np.maximum(eps, prior.output_probabilities)
            # renormalize B to make it row-stochastic
            full_obs_probabilities /= full_obs_probabilities.sum(axis=1)[:, None]
        else:
            full_obs_probabilities = prior.output_probabilities

        maxT = max(len(o) for o in dtrajs_lagged_strided)

        # pre-construct hidden variables
        temp_alpha = np.zeros((maxT, prior.n_hidden_states))

        has_all_obs_symbols = model.prior.n_observation_states == len(model.prior.observation_symbols_full)

        try:
            # sample model is basically copy of prior
            sample_model = BayesianHMM._SampleStorage(
                transition_matrix=prior.transition_model.transition_matrix.copy(),
                output_model=DiscreteOutputModel(full_obs_probabilities.copy()),
                initial_distribution=prior.initial_distribution.copy(),
                stationary_distribution=prior.transition_model.stationary_distribution.copy(),
                counts=prior.count_model.count_matrix.copy(),
                hidden_trajs=[]
            )

            # Run burn-in.
            for _ in range(n_burn_in):
                self._update(sample_model, dtrajs_lagged_strided, temp_alpha, transition_matrix_prior,
                             initial_distribution_prior)

            # Collect data.
            models = []
            for _ in progress(range(self.n_samples), desc="Drawing samples", leave=False):
                # Run a number of Gibbs sampling updates to generate each sample.
                for _ in range(n_thin):
                    self._update(sample_model, dtrajs_lagged_strided, temp_alpha, transition_matrix_prior,
                                 initial_distribution_prior)
                    sample_model.output_model.normalize()
                self._append_sample(models, prior, sample_model)

            if not has_all_obs_symbols:
                models = [m.submodel(states=None, obs=model.prior.observation_symbols) for m in models]

            model.samples = models
        finally:
            del temp_alpha

        # set new model
        self._model = model

        return self

    def _append_sample(self, models, prior, sample_model):
        # Save a copy of the current model.
        model_copy = deepcopy(sample_model)
        # the Viterbi path is discarded, but is needed to get a new transition matrix for each model.
        if not self.store_hidden:
            model_copy.hidden_trajs.clear()
        # potentially restrict sampled models to observed space
        # since model_copy is defined on full space, observation_symbols are also observation states
        count_model = TransitionCountModel(model_copy.counts, lagtime=prior.lagtime)
        models.append(HiddenMarkovModel(
            transition_model=MarkovStateModel(model_copy.transition_matrix,
                                              stationary_distribution=model_copy.stationary_distribution,
                                              reversible=self.reversible, count_model=count_model),
            output_model=model_copy.output_model, initial_distribution=model_copy.initial_distribution,
            hidden_state_trajectories=model_copy.hidden_trajs))

    def chapman_kolmogorov_validator(self, mlags, test_model: BayesianHMMPosterior = None):
        r"""Returns a Chapman-Kolmogorov validator based on this estimator and a test model.

        Parameters
        ----------
        mlags : int or int-array
            Multiple of lagtimes of the test_model to test against.
        test_model : BayesianHMMPosterior, optional, default=None
            The model that is tested. If not provided, uses this estimator's encapsulated model.

        Returns
        -------
        validator : markov.MembershipsChapmanKolmogorovValidator
            The validator.
        """
        test_model = self.fetch_model() if test_model is None else test_model
        assert test_model is not None, "We need a test model via argument or an estimator which was already" \
                                       "fit to data."

        from . import DiscreteOutputModel
        assert isinstance(test_model.prior.output_model, DiscreteOutputModel), \
            "Can only perform CKTest for discrete output models"
        memberships = np.eye(test_model.prior.n_hidden_states)
        lagtime = test_model.prior.lagtime
        return BayesianHMMChapmanKolmogorovValidator(test_model, self, memberships, lagtime, mlags)


def _ck_estimate_model_for_lag(estimator: BayesianHMM, model, data, lagtime):
    estimator = BayesianHMM.default(dtrajs=data, n_hidden_states=estimator.initial_hmm.n_hidden_states,
                                    lagtime=lagtime, n_samples=estimator.n_samples, stride=estimator.stride,
                                    initial_distribution_prior=estimator.initial_distribution_prior,
                                    transition_matrix_prior=estimator.transition_matrix_prior,
                                    reversible=estimator.reversible, stationary=estimator.stationary)
    return estimator.fit(data).fetch_model()


class BayesianHMMChapmanKolmogorovValidator(MembershipsChapmanKolmogorovValidator):

    def fit(self, data, n_jobs=1, progress=None, **kw):
        if n_jobs != 1:
            import warnings
            warnings.warn("ignoring n_jobs for HMM CKtest")

        return super().fit(data, n_jobs=1, estimate_model_for_lag=_ck_estimate_model_for_lag, progress=progress, **kw)



r"""
.. currentmodule:: deeptime.markov.tools

MSM functions
=============
Low-level functions for estimation and analysis of transition matrices and io.
"""


from . import analysis
from . import estimation
from . import flux



r"""This module contains sparse implementation of the fingerprint module

.. moduleauthor:: B.Trendelkamp-Schroer <benjamin DOT trendelkamp-schroer AT fu-berlin DOT de>

"""

import numpy as np

from ._decomposition import rdl_decomposition, timescales_from_eigenvalues
from ._decomposition import stationary_distribution as statdist

################################################################################
# Fingerprints
################################################################################


def fingerprint_correlation(P, obs1, obs2=None, tau=1, k=None, ncv=None):
    r"""Compute dynamical fingerprint crosscorrelation.

    The dynamical fingerprint autocorrelation is the timescale
    amplitude spectrum of the autocorrelation of the given observables
    under the action of the dynamics P

    Parameters
    ----------
    P : ndarray, shape=(n, n) or scipy.sparse matrix
        Transition matrix
    obs1 : ndarray, shape=(n,)
        Vector representing observable 1 on discrete states
    obs2 : ndarray, shape=(n,)
        Vector representing observable 2 on discrete states.
        If none, obs2=obs1, i.e. the autocorrelation is used
    tau : lag time of the the transition matrix. Used for
        computing the timescales returned
    k : int (optional)
        Number of amplitudes
    ncv : int (optional)
        The number of Lanczos vectors generated, `ncv` must be greater than k;
        it is recommended that ncv > 2*k


    Returns
    -------
    timescales : (N,) ndarray
        Time-scales of the transition matrix
    amplitudes : (N,) ndarray
        Amplitudes for the correlation experiment

    """
    return fingerprint(P, obs1, obs2=obs2, tau=tau, k=k, ncv=ncv)


def fingerprint_relaxation(P, p0, obs, tau=1, k=None, ncv=None):
    r"""Compute dynamical fingerprint crosscorrelation.

    The dynamical fingerprint autocorrelation is the timescale
    amplitude spectrum of the autocorrelation of the given observables
    under the action of the dynamics P

    Parameters
    ----------
    P : ndarray, shape=(n, n) or scipy.sparse matrix
        Transition matrix
    p0 : ndarray, shape=(n)
        starting distribution
    obs : ndarray, shape=(n)
        Vector representing observable 2 on discrete states.
        If none, obs2=obs1, i.e. the autocorrelation is used
    tau : lag time of the the transition matrix. Used for
        computing the timescales returned
    k : int (optional)
        Number of amplitudes
    ncv : int (optional)
        The number of Lanczos vectors generated, `ncv` must be greater than k;
        it is recommended that ncv > 2*k

    Returns
    -------
    timescales : (N,) ndarray
        Time-scales of the transition matrix
    amplitudes : (N,) ndarray
        Amplitudes for the relaxation experiment

    """
    one_vec = np.ones(P.shape[0])
    return fingerprint(P, one_vec, obs2=obs, p0=p0, tau=tau, k=k, ncv=ncv)


def fingerprint(P, obs1, obs2=None, p0=None, tau=1, k=None, ncv=None):
    r"""Dynamical fingerprint for equilibrium or relaxation experiment

    The dynamical fingerprint is given by the implied time-scale
    spectrum together with the corresponding amplitudes.

    Parameters
    ----------
    P : (M, M) scipy.sparse matrix
        Transition matrix
    obs1 : (M,) ndarray
        Observable, represented as vector on state space
    obs2 : (M,) ndarray (optional)
        Second observable, for cross-correlations
    p0 : (M,) ndarray (optional)
        Initial distribution for a relaxation experiment
    tau : int (optional)
        Lag time of given transition matrix, for correct time-scales
    k : int (optional)
        Number of time-scales and amplitudes to compute
    ncv : int (optional)
        The number of Lanczos vectors generated, `ncv` must be greater than k;
        it is recommended that ncv > 2*k


    Returns
    -------
    timescales : (N,) ndarray
        Time-scales of the transition matrix
    amplitudes : (N,) ndarray
        Amplitudes for the given observable(s)

    """
    if obs2 is None:
        obs2 = obs1
    R, D, L = rdl_decomposition(P, k=k, ncv=ncv)
    """Stationary vector"""
    mu = L[0, :]
    """Extract diagonal"""
    w = np.diagonal(D)
    """Compute time-scales"""
    timescales = timescales_from_eigenvalues(w, tau)
    if p0 is None:
        """Use stationary distribution - we can not use only left
        eigenvectors since the system might be non-reversible"""
        amplitudes = np.dot(mu * obs1, R) * np.dot(L, obs2)
    else:
        """Use initial distribution"""
        amplitudes = np.dot(p0 * obs1, R) * np.dot(L, obs2)
    return timescales, amplitudes


################################################################################
# Correlation
################################################################################

def correlation(P, obs1, obs2=None, times=(1,), k=None, ncv=None):
    r"""Time-correlation for equilibrium experiment.

    Parameters
    ----------
    P : (M, M) ndarray
        Transition matrix
    obs1 : (M,) ndarray
        Observable, represented as vector on state space
    obs2 : (M,) ndarray (optional)
        Second observable, for cross-correlations
    times : list or tuple of int (optional)
        List of times (in tau) at which to compute correlation
    k : int (optional)
        Number of eigenvalues and amplitudes to use for computation
    ncv : int (optional)
        The number of Lanczos vectors generated, `ncv` must be greater than k;
        it is recommended that ncv > 2*k

    Returns
    -------
    correlations : ndarray
        Correlation values at given times

    """
    M = P.shape[0]
    T = np.asarray(times).max()
    if T < M:
        return correlation_matvec(P, obs1, obs2=obs2, times=times)
    else:
        return correlation_decomp(P, obs1, obs2=obs2, times=times, k=k, ncv=ncv)


def correlation_decomp(P, obs1, obs2=None, times=(1,), k=None, ncv=None):
    r"""Time-correlation for equilibrium experiment - via decomposition.

    Parameters
    ----------
    P : (M, M) ndarray
        Transition matrix
    obs1 : (M,) ndarray
        Observable, represented as vector on state space
    obs2 : (M,) ndarray (optional)
        Second observable, for cross-correlations
    times : list or tuple of int (optional)
        List of times (in tau) at which to compute correlation
    k : int (optional)
        Number of eigenvalues and amplitudes to use for computation
    ncv : int (optional)
        The number of Lanczos vectors generated, `ncv` must be greater than k;
        it is recommended that ncv > 2*k

    Returns
    -------
    correlations : ndarray
        Correlation values at given times

    """
    if obs2 is None:
        obs2 = obs1
    R, D, L = rdl_decomposition(P, k=k, ncv=ncv)
    """Stationary vector"""
    mu = L[0, :]
    """Extract eigenvalues"""
    ev = np.diagonal(D)
    """Amplitudes"""
    amplitudes = np.dot(mu * obs1, R) * np.dot(L, obs2)
    """Propgate eigenvalues"""
    times = np.asarray(times)
    ev_t = ev[np.newaxis, :] ** times[:, np.newaxis]
    """Compute result"""
    res = np.dot(ev_t, amplitudes)
    """Truncate imaginary part - should be zero anyways"""
    res = res.real
    return res


def correlation_matvec(P, obs1, obs2=None, times=(1,)):
    r"""Time-correlation for equilibrium experiment - via matrix vector products.

    Parameters
    ----------
    P : (M, M) ndarray
        Transition matrix
    obs1 : (M,) ndarray
        Observable, represented as vector on state space
    obs2 : (M,) ndarray (optional)
        Second observable, for cross-correlations
    times : list or tuple of int (optional)
        List of times (in tau) at which to compute correlation

    Returns
    -------
    correlations : ndarray
        Correlation values at given times

    """
    if obs2 is None:
        obs2 = obs1

    """Compute stationary vector"""
    mu = statdist(P)
    obs1mu = mu * obs1

    times = np.asarray(times)
    """Sort in increasing order"""
    ind = np.argsort(times)
    times = times[ind]

    if times[0] < 0:
        raise ValueError("Times can not be negative")
    dt = times[1:] - times[0:-1]

    nt = len(times)

    correlations = np.zeros(nt)

    """Propagate obs2 to initial time"""
    obs2_t = 1.0 * obs2
    obs2_t = propagate(P, obs2_t, times[0])
    correlations[0] = np.dot(obs1mu, obs2_t)
    for i in range(nt - 1):
        obs2_t = propagate(P, obs2_t, dt[i])
        correlations[i + 1] = np.dot(obs1mu, obs2_t)

    """Cast back to original order of time points"""
    correlations = correlations[ind]

    return correlations


################################################################################
# Relaxation
################################################################################


def relaxation(P, p0, obs, times=(1,), k=None, ncv=None):
    r"""Relaxation experiment.

    The relaxation experiment describes the time-evolution
    of an expectation value starting in a non-equilibrium
    situation.

    Parameters
    ----------
    P : (M, M) ndarray
        Transition matrix
    p0 : (M,) ndarray (optional)
        Initial distribution for a relaxation experiment
    obs : (M,) ndarray
        Observable, represented as vector on state space
    times : list of int (optional)
        List of times at which to compute expectation
    k : int (optional)
        Number of eigenvalues and amplitudes to use for computation
    ncv : int (optional)
        The number of Lanczos vectors generated, `ncv` must be greater than k;
        it is recommended that ncv > 2*k


    Returns
    -------
    res : ndarray
        Array of expectation value at given times

    """
    M = P.shape[0]
    T = np.asarray(times).max()
    if T < M:
        return relaxation_matvec(P, p0, obs, times=times)
    else:
        return relaxation_decomp(P, p0, obs, times=times, k=k, ncv=ncv)


def relaxation_decomp(P, p0, obs, times=(1,), k=None, ncv=None):
    r"""Relaxation experiment.

    The relaxation experiment describes the time-evolution
    of an expectation value starting in a non-equilibrium
    situation.

    Parameters
    ----------
    P : (M, M) ndarray
        Transition matrix
    p0 : (M,) ndarray (optional)
        Initial distribution for a relaxation experiment
    obs : (M,) ndarray
        Observable, represented as vector on state space
    times : list of int (optional)
        List of times at which to compute expectation
    k : int (optional)
        Number of eigenvalues and amplitudes to use for computation
    ncv : int (optional)
        The number of Lanczos vectors generated, `ncv` must be greater than k;
        it is recommended that ncv > 2*k

    Returns
    -------
    res : ndarray
        Array of expectation value at given times

    """
    R, D, L = rdl_decomposition(P, k=k, ncv=ncv)
    """Extract eigenvalues"""
    ev = np.diagonal(D)
    """Amplitudes"""
    amplitudes = np.dot(p0, R) * np.dot(L, obs)
    """Propgate eigenvalues"""
    times = np.asarray(times)
    ev_t = ev[np.newaxis, :] ** times[:, np.newaxis]
    """Compute result"""
    res = np.dot(ev_t, amplitudes)
    """Truncate imgainary part - is zero anyways"""
    res = res.real
    return res


def relaxation_matvec(P, p0, obs, times=(1,)):
    r"""Relaxation experiment.

    The relaxation experiment describes the time-evolution
    of an expectation value starting in a non-equilibrium
    situation.

    Parameters
    ----------
    P : (M, M) ndarray
        Transition matrix
    p0 : (M,) ndarray (optional)
        Initial distribution for a relaxation experiment
    obs : (M,) ndarray
        Observable, represented as vector on state space
    times : list of int (optional)
        List of times at which to compute expectation

    Returns
    -------
    res : ndarray
        Array of expectation value at given times

    """
    times = np.asarray(times)
    """Sort in increasing order"""
    ind = np.argsort(times)
    times = times[ind]

    if times[0] < 0:
        raise ValueError("Times can not be negative")
    dt = times[1:] - times[0:-1]

    nt = len(times)

    relaxations = np.zeros(nt)

    """Propagate obs to initial time"""
    obs_t = 1.0 * obs
    obs_t = propagate(P, obs_t, times[0])
    relaxations[0] = np.dot(p0, obs_t)
    for i in range(nt - 1):
        obs_t = propagate(P, obs_t, dt[i])
        relaxations[i + 1] = np.dot(p0, obs_t)

    """Cast back to original order of time points"""
    relaxations = relaxations[ind]

    return relaxations


################################################################################
# Helper functions
################################################################################

def propagate(A, x, N):
    r"""Use matrix A to propagate vector x.

    Parameters
    ----------
    A : (M, M) scipy.sparse matrix
        Matrix of propagator
    x : (M, ) ndarray or scipy.sparse matrix
        Vector to propagate
    N : int
        Number of steps to propagate

    Returns
    -------
    y : (M, ) ndarray or scipy.sparse matrix
        Propagated vector

    """
    y = 1.0 * x
    for i in range(N):
        y = A.dot(y)
    return y



"""
Created on Jan 13, 2014

@author: noe
"""

import numpy as np


def transition_matrix_non_reversible(C):
    r"""
    Estimates a non-reversible transition matrix from count matrix C

    T_ij = c_ij / c_i where c_i = sum_j c_ij

    Parameters
    ----------
    C: ndarray, shape (n,n)
        count matrix

    Returns
    -------
    T: Estimated transition matrix

    """
    # multiply by 1.0 to make sure we're not doing integer division
    rowsums = 1.0 * np.sum(C, axis=1)
    if np.min(rowsums) <= 0:
        raise ValueError(
            "Transition matrix has row sum of " + str(np.min(rowsums)) + ". Must have strictly positive row sums.")
    return np.divide(C, rowsums[:, np.newaxis])


def transition_matrix_reversible_pisym(C, return_statdist=False):
    r"""
    Estimates reversible transition matrix as follows:

    ..:math:
        p_{ij} = c_{ij} / c_i where c_i = sum_j c_{ij}
        \pi_j = \sum_j \pi_i p_{ij}
        x_{ij} = \pi_i p_{ij} + \pi_j p_{ji}
        p^{rev}_{ij} = x_{ij} / x_i where x_i = sum_j x_{ij}

    In words: takes the nonreversible transition matrix estimate, uses its
    stationary distribution to compute an equilibrium correlation matrix,
    symmetrizes that correlation matrix and then normalizes to the reversible
    transition matrix estimate.

    Parameters
    ----------
    C: ndarray, shape (n,n)
        count matrix

    Returns
    -------
    T: Estimated transition matrix

    """
    # nonreversible estimate
    T_nonrev = transition_matrix_non_reversible(C)
    from ...analysis import stationary_distribution
    pi = stationary_distribution(T_nonrev)
    # correlation matrix
    X = pi[:, None] * T_nonrev
    X = X.T + X
    # result
    T_rev = X / X.sum(axis=1)[:, None]
    if return_statdist:
        return T_rev, pi
    return T_rev



from math import sqrt
from typing import Optional, Callable

import numpy as np

from ...base import Estimator
from ...numeric import is_square_matrix
from .. import TransitionCountEstimator
from .._base import _MSMBaseEstimator, BayesianPosterior, MembershipsChapmanKolmogorovValidator
from . import MarkovStateModel, MaximumLikelihoodMSM

__author__ = 'noe, marscher, clonker'


class BayesianMSM(_MSMBaseEstimator):
    r""" Bayesian estimator for MSMs given discrete trajectory statistics.

    Implementation following :footcite:`trendelkamp2015estimation`.

    Parameters
    ----------
    n_samples : int, optional, default=100
        Number of sampled transition matrices used in estimation of confidences.
    n_steps : int, optional, default=None
        Number of Gibbs sampling steps for each transition matrix. If None, nsteps will be determined
        automatically as the square root of the number of states in the full state space of the count matrix.
        This is a heuristic for the number of steps it takes to decorrelate between samples.
    reversible : bool, optional, default=True
        If true compute reversible MSM, else non-reversible MSM.
    stationary_distribution_constraint : ndarray, optional, default=None
        Stationary vector on the full set of states. Assign zero stationary probabilities to states for which the
        stationary vector is unknown. Estimation will be made such that the resulting ensemble of transition
        matrices is defined on the intersection of the states with positive stationary vector and the largest
        connected set (undirected in the default case).
    sparse : bool, optional, default=False
        If true compute count matrix, transition matrix and all derived quantities using sparse matrix algebra. In
        this case python sparse matrices will be returned by the corresponding functions instead of numpy arrays.
        This behavior is suggested for very large numbers of states (e.g. > 4000) because it is likely to be much
        more efficient.
    confidence : float, optional, default=0.954
        Confidence interval. By default two sigma (95.4%) is used. Use 68.3% for one sigma, 99.7% for three sigma.
    maxiter : int, optional, default=1000000
        Optional parameter with reversible = True, sets the maximum number of iterations before the transition
        matrix estimation method exits.
    maxerr : float, optional, default = 1e-8
        Optional parameter with reversible = True. Convergence tolerance for transition matrix estimation. This
        specifies the maximum change of the Euclidean norm of relative stationary probabilities
        (:math:`x_i = \sum_k x_{ik}`). The relative stationary probability changes
        :math:`e_i = (x_i^{(1)} - x_i^{(2)})/(x_i^{(1)} + x_i^{(2)})` are used in order to track changes in small
        probabilities. The Euclidean norm of the change vector, :math:`|e_i|_2`, is compared to maxerr.

    References
    ----------
    .. footbibliography::

    Examples
    --------
    Note that the following example is only qualitatively and not
    quantitatively reproducible because it involves random numbers.

    We build a Bayesian Markov model for the following two trajectories at lag
    time :math:`\tau=2`:

    >>> import numpy as np
    >>> import deeptime
    >>> dtrajs = [np.array([0,1,2,2,2,2,1,2,2,2,1,0,0,0,0,0,0,0]), np.array([0,0,0,0,1,1,2,2,2,2,2,2,2,1,0,0])]
    >>> counts = deeptime.markov.TransitionCountEstimator(lagtime=2, count_mode="effective").fit(dtrajs).fetch_model()
    >>> mm = deeptime.markov.msm.BayesianMSM().fit(counts).fetch_model()

    The resulting Model contains a prior :class:`MSM <deeptime.markov.msm.MarkovStateModel>` as well as a list of sample
    MSMs. Its transition matrix comes from a maximum likelihood estimation. We can access, e.g., the transition matrix
    as follows:

    >>> print(mm.prior.transition_matrix)  # doctest: +SKIP
    [[ 0.70000001  0.16463699  0.135363  ]
     [ 0.38169055  0.          0.61830945]
     [ 0.12023989  0.23690297  0.64285714]]

    Furthermore the bayesian MSM posterior returned by :meth:`fetch_model` is able to
    compute the probability distribution and statistical models of all methods
    that are offered by the MSM object. This works as follows. The :meth:`BayesianPosterior.gather_stats` method
    takes as argument the method you want to evaluate and then returns a statistics summary over requested quantity:

    >>> print(mm.gather_stats('transition_matrix').mean)  # doctest: +SKIP
    [[ 0.71108663  0.15947371  0.12943966]
     [ 0.41076105  0.          0.58923895]
     [ 0.13079372  0.23005443  0.63915185]]

    Likewise, the standard deviation by element:

    >>> print(mm.gather_stats('transition_matrix').std)  # doctest: +SKIP
    [[ 0.13707029  0.09479627  0.09200214]
     [ 0.15247454  0.          0.15247454]
     [ 0.07701315  0.09385258  0.1119089 ]]

    And this is the 95% (2 sigma) confidence interval. You can control the
    percentile using the conf argument in this function:

    >>> stats = mm.gather_stats('transition_matrix')
    >>> print(stats.L) # doctest: +SKIP
    >>> print(stats.R)  # doctest: +SKIP
    [[ 0.44083423  0.03926518  0.0242113 ]
     [ 0.14102544  0.          0.30729828]
     [ 0.02440188  0.07629456  0.43682481]]
    [[ 0.93571706  0.37522581  0.40180041]
     [ 0.69307665  0.          0.8649215 ]
     [ 0.31029752  0.44035732  0.85994006]]

    If you want to compute expectations of functions that require arguments,
    just pass these arguments as well:

    >>> print(mm.gather_stats('mfpt', A=0, B=2)) # doctest: +SKIP
    12.9049811296

    And if you want to histogram the distribution or compute more complex
    statistical moment such as the covariance between different quantities,
    just get the full sample of your quantity of interest and evaluate it
    at will:

    >>> samples = mm.gather_stats('mfpt', store_samples=True, A=0, B=2).samples
    >>> print(samples[:4]) # doctest: +SKIP
    [7.9763615793248155, 8.6540958274695701, 26.295326015231058, 17.909895469938899]

    Internally, the SampledMSM object has 100 transition matrices (the number
    can be controlled by nsamples), that were computed by the transition matrix
    sampling method. All of the above sample functions iterate over these 100
    transition matrices and evaluate the requested function with the given
    parameters on each of them.
    """

    def __init__(self, n_samples: int = 100, n_steps: int = None, reversible: bool = True,
                 stationary_distribution_constraint: Optional[np.ndarray] = None,
                 sparse: bool = False, confidence: float = 0.954, maxiter: int = int(1e6), maxerr: float = 1e-8):
        super(BayesianMSM, self).__init__(reversible=reversible, sparse=sparse)
        self.stationary_distribution_constraint = stationary_distribution_constraint
        self.maxiter = maxiter
        self.maxerr = maxerr
        self.n_samples = n_samples
        self.n_steps = n_steps
        self.confidence = confidence

    @property
    def stationary_distribution_constraint(self) -> Optional[np.ndarray]:
        r"""
        The stationary distribution constraint that can either be None (no constraint) or constrains the
        count and transition matrices to states with positive stationary vector entries.

        :getter: Retrieves the currently configured constraint, can be None.
        :setter: Sets a stationary distribution constraint by giving a stationary vector as value. The estimated count-
                 and transition-matrices are restricted to states that have positive entries. In case the vector is
                 not normalized, setting it here implicitly copies and normalizes it.
        :type: ndarray or None
        """
        return self._stationary_distribution_constraint

    @stationary_distribution_constraint.setter
    def stationary_distribution_constraint(self, value: Optional[np.ndarray]):
        if value is not None and (np.any(value < 0) or np.any(value > 1)):
            raise ValueError("not a distribution, contained negative entries and/or entries > 1.")
        if value is not None and np.sum(value) != 1.0:
            # re-normalize if not already normalized
            value = np.copy(value) / np.sum(value)
        self._stationary_distribution_constraint = value

    def fetch_model(self) -> Optional[BayesianPosterior]:
        r"""
        Yields the model that was estimated the most recent.

        Returns
        -------
        model : BayesianPosterior or None
            The estimated model or None if fit was not called.
        """
        return self._model

    def fit(self, data, callback: Callable = None):
        """
        Performs the estimation on either a count matrix or a previously estimated TransitionCountModel.

        Parameters
        ----------
        data : (N,N) count matrix or TransitionCountModel or MaximumLikelihoodMSM or MarkovStateModel
            a count matrix or a transition count model that was estimated from data

        callback: callable, optional, default=None
            Function to be called to indicate progress of sampling.

        Returns
        -------
        self : BayesianMSM
            Reference to self.
        """
        from deeptime.markov import TransitionCountModel
        if isinstance(data, TransitionCountModel) and data.counting_mode is not None \
                and "effective" not in data.counting_mode:
            raise ValueError("The transition count model was not estimated using an effective counting method, "
                             "therefore counts are likely to be strongly correlated yielding wrong confidences.")

        if isinstance(data, Estimator):
            if data.has_model:
                data = data.fetch_model()
            else:
                raise ValueError("Can only use estimators as input if they have been fit previously.")

        if isinstance(data, TransitionCountModel) or is_square_matrix(data):
            msm = MaximumLikelihoodMSM(
                reversible=self.reversible, stationary_distribution_constraint=self.stationary_distribution_constraint,
                sparse=self.sparse, maxiter=self.maxiter, maxerr=self.maxerr
            ).fit(data).fetch_model()
        elif isinstance(data, MarkovStateModel):
            msm = data
        else:
            raise ValueError("Unsupported input data, can only be count matrix (or TransitionCountModel, "
                             "TransitionCountEstimator) or a MarkovStateModel instance or an estimator producing "
                             "Markov state models.")

        return self.fit_from_msm(msm, callback=callback)

    def sample(self, prior: MarkovStateModel, n_samples: int, n_steps: Optional[int] = None, callback=None):
        r""" Performs sampling based on a prior.

        Parameters
        ----------
        prior : MarkovStateModel
            The MSM that is used as initial sampling point.
        n_samples : int
            The number of samples to draw.
        n_steps : int, optional, default=None
            The number of sampling steps for each transition matrix. If None, determined
            by :math:`\sqrt{\mathrm{n\_states}}`.
        callback : callable, optional, default=None
            Callback function that indicates progress of sampling.

        Returns
        -------
        samples : list of :obj:`MarkovStateModel`
            The generated samples

        Examples
        --------
        This method can in particular be used to append samples to an already estimated posterior:

        >>> import numpy as np
        >>> import deeptime as dt
        >>> dtrajs = [np.array([0,1,2,2,2,2,1,2,2,2,1,0,0,0,0,0,0,0]),
        ...           np.array([0,0,0,0,1,1,2,2,2,2,2,2,2,1,0,0])]
        >>> prior = dt.markov.msm.MaximumLikelihoodMSM().fit(dtrajs, lagtime=1)
        >>> estimator = dt.markov.msm.BayesianMSM()
        >>> posterior = estimator.fit(prior).fetch_model()
        >>> n_samples = len(posterior.samples)
        >>> posterior.samples.extend(estimator.sample(posterior.prior, n_samples=23))
        >>> assert len(posterior.samples) == n_samples + 23
        """
        if n_steps is None:
            # heuristic for number of steps to decorrelate
            n_steps = int(sqrt(prior.count_model.n_states_full))
        # transition matrix sampler
        from deeptime.markov.tools.estimation import tmatrix_sampler
        if self.stationary_distribution_constraint is None:
            tsampler = tmatrix_sampler(prior.count_model.count_matrix, reversible=self.reversible,
                                       T0=prior.transition_matrix, nsteps=n_steps)
        else:
            # Use the stationary distribution on the active set of states
            statdist_active = prior.stationary_distribution
            # We can not use the MLE as T0. Use the initialization in the reversible pi sampler
            tsampler = tmatrix_sampler(prior.count_model.count_matrix, reversible=self.reversible,
                                       mu=statdist_active, nsteps=n_steps)
        sample_Ps, sample_mus = tsampler.sample(nsamples=n_samples, return_statdist=True, callback=callback)
        # construct sampled MSMs
        samples = [
            MarkovStateModel(P, stationary_distribution=pi, reversible=self.reversible,
                             count_model=prior.count_model,
                             transition_matrix_tolerance=prior.transition_matrix_tolerance)
            for P, pi in zip(sample_Ps, sample_mus)
        ]
        return samples

    def fit_from_msm(self, msm: MarkovStateModel, callback=None):
        r""" Fits a bayesian posterior from a given Markov state model. The MSM must contain a count model to be able
        to produce confidences. Note that the count model should be produced using effective counting, otherwise
        counts are correlated and computed confidences are wrong.

        Parameters
        ----------
        msm : MarkovStateModel
            The Markov state model to use as sampling start point.
        callback : callable, optional, default=None
            Function to be called to indicate progress of sampling.

        Returns
        -------
        self : BayesianMSM
            Reference to self.
        """
        if not msm.has_count_model:
            raise ValueError("Can only sample confidences with a count model. The counting mode should be 'effective'"
                             " to avoid correlations between counts and therefore wrong confidences.")
        # use the same count matrix as the MLE. This is why we have effective as a default
        samples = self.sample(msm, self.n_samples, self.n_steps, callback)
        self._model = BayesianPosterior(prior=msm, samples=samples)
        return self

    def chapman_kolmogorov_validator(self, n_metastable_sets: int, mlags, test_model: BayesianPosterior = None):
        r"""Returns a Chapman-Kolmogorov validator based on this estimator and a test model.

        Parameters
        ----------
        n_metastable_sets : int
            Number of metastable sets to project the state space down to.
        mlags : int or range or list
            Multiple of lagtimes of the test_model to test against.
        test_model : BayesianPosterior, optional, default=None
            The model that is tested. If not provided, uses this estimator's encapsulated model.

        Returns
        -------
        validator : markov.MembershipsChapmanKolmogorovValidator
            The validator.
        """
        test_model = self.fetch_model() if test_model is None else test_model
        assert test_model is not None, "We need a test model via argument or an estimator which was already" \
                                       "fit to data."
        prior = test_model.prior
        assert prior.has_count_model, "The test model needs to have a count model, i.e., be estimated from data."
        pcca = prior.pcca(n_metastable_sets)
        reference_lagtime = prior.count_model.lagtime
        return BayesianMSMChapmanKolmogorovValidator(test_model, self, pcca.memberships, reference_lagtime, mlags)


def _ck_estimate_model_for_lag(estimator, model, data, lagtime):
    counting_mode = model.prior.count_model.counting_mode
    counts = TransitionCountEstimator(lagtime, counting_mode).fit(data, n_jobs=1).fetch_model().submodel_largest()
    return estimator.fit(counts).fetch_model()


class BayesianMSMChapmanKolmogorovValidator(MembershipsChapmanKolmogorovValidator):

    def fit(self, data, n_jobs=None, progress=None, **kw):
        return super().fit(data, n_jobs, progress, estimate_model_for_lag=_ck_estimate_model_for_lag, **kw)


import numpy as np
from scipy.spatial import distance

from . import Kernel


class GaussianKernel(Kernel):
    r""" Implementation of the Gaussian kernel

    .. math::

        \kappa (x,y) = \exp \left(-\sum_k (x_k - y_k)^2 / (2\sigma^2)\right),

    where :math:`\sigma` is the bandwidth of the kernel.

    Parameters
    ----------
    sigma : float
        The bandwidth.
    """

    valid_impls = 'cdist', 'binomial'  #: Valid implementation modes.

    def __init__(self, sigma, impl='cdist'):
        assert impl in GaussianKernel.valid_impls, f"impl needs to be one of {GaussianKernel.valid_impls}"
        self._sigma = np.array(sigma, dtype=np.float64)
        self.impl = impl

    @property
    def impl(self):
        return self._impl

    @impl.setter
    def impl(self, value):
        assert value in GaussianKernel.valid_impls, f"impl {value} not known, " \
                                                    f"supported are {GaussianKernel.valid_impls}"
        self._impl = value

    @property
    def sigma(self) -> np.ndarray:
        r""" Bandwidth of the Gaussian kernel.

        :getter: Yields the bandwidth.
        :type: (1,) ndarray
        """
        return self._sigma

    def _evaluate(self, x, y) -> float:
        return np.exp(-np.square(np.linalg.norm(x - y)) / (2 * np.square(self.sigma)))

    def apply(self, data_1: np.ndarray, data_2: np.ndarray) -> np.ndarray:
        if self.impl == 'cdist':
            D = distance.cdist(data_1, data_2, metric='sqeuclidean')
        elif self.impl == 'binomial':
            x1_norm = np.square(data_1).sum(axis=-1, keepdims=True)
            x2_norm = np.square(data_2).sum(axis=-1, keepdims=True)
            D = x2_norm.T - 2. * data_1 @ data_2.T + x1_norm
            D = np.clip(D, a_min=1e-16, a_max=None)
        return np.exp(-D / (2 * self.sigma * self.sigma))

    def __str__(self):
        return f"GaussianKernel[sigma={self.sigma}, impl={self.impl}]"


class GeneralizedGaussianKernel(Kernel):
    r""" Implementation of the generalized Gaussian kernel with bandwidth per dimension. It is defined by

    .. math::

        \kappa (x,y) = \exp \left( \frac{1}{2}(x-y)^\top \Sigma^{-1} (x-y) \right),

    where :math:`\Sigma = \mathrm{diag} (\sigma_1,\ldots,\sigma_d)` are the bandwidths of the kernel.

    Parameters
    ----------
    sigmas : (d,) ndarray
        The bandwidths. Must match the dimension of the data that is used to evaluate the kernel.
    """

    def __init__(self, sigmas):
        self._sigmas = sigmas.squeeze()
        self._D = np.diag(1. / (2 * np.square(sigmas)))
        self._sqrt_D = 1. / (np.sqrt(2) * sigmas)

    def _evaluate(self, x, y) -> float:
        diff = (x - y).squeeze()
        return np.exp(-diff @ self._D @ diff.T)

    def apply(self, data_1: np.ndarray, data_2: np.ndarray) -> np.ndarray:
        ri = np.expand_dims(data_1, axis=1)
        rj = np.expand_dims(data_2, axis=0)
        rij = (ri - rj) * self._sqrt_D
        D = np.add.reduce(np.square(rij), axis=-1, keepdims=False)
        return np.exp(-D)

    def __str__(self):
        return f"GeneralizedGaussianKernel[sigmas={','.join(f'{s:.3f}' for s in self._sigmas)}]"


class LaplacianKernel(Kernel):
    r""" Implementation of the Laplacian kernel

    .. math::

        \kappa (x,y) = \exp \left( \| x-y \|_2 / \sigma\right),

    where :math:`\sigma` is the bandwidth of the kernel."""

    def __init__(self, sigma):
        self._sigma = sigma

    def _evaluate(self, x, y) -> float:
        return np.exp(-np.linalg.norm(x - y) / self._sigma)

    def apply(self, data_1: np.ndarray, data_2: np.ndarray) -> np.ndarray:
        return np.exp(-distance.cdist(data_1, data_2, metric='euclidean') / self._sigma)

    def __str__(self):
        return f"LaplacianKernel[sigma={self._sigma:.3f}]"


class PolynomialKernel(Kernel):
    r""" Implementation of the polynomial kernel

    .. math::

        \kappa (x,y) = (x^\top y + c)^d,

    where :math:`p` is the degree and :math:`c` is the inhomogeneity.

    Parameters
    ----------
    degree : int
        The degree, must be non-negative.
    inhomogeneity : float, must be non-negative, optional, default=1.
        The inhomogeneity.
    """

    def __init__(self, degree: int, inhomogeneity: float = 1.):
        assert degree >= 0
        assert inhomogeneity >= 0
        self.degree = degree
        self.inhomogeneity = inhomogeneity

    def _evaluate(self, x, y) -> float:
        return (self.inhomogeneity + np.dot(x, y)) ** self.degree

    def apply(self, data_1: np.ndarray, data_2: np.ndarray) -> np.ndarray:
        ri = np.expand_dims(data_1, axis=1)
        rj = np.expand_dims(data_2, axis=0)
        scalar_products = np.add.reduce(ri * rj, axis=-1, keepdims=False)
        return (self.inhomogeneity + scalar_products) ** self.degree

    def __str__(self):
        return f"PolynomialKernel[degree={self.degree}, inhomogeneity={self.inhomogeneity}]"



r"""
.. currentmodule: deeptime.numeric

===============================================================================
General numerical tools
===============================================================================

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst
    
    drop_nan_rows
    is_diagonal_matrix
    is_square_matrix
    schatten_norm

===============================================================================
Numerical tools for eigenvalue problems
===============================================================================

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    eigs
    eig_corr
    sort_eigs
    spd_eig
    spd_truncated_svd
    spd_inv
    spd_inv_split
    spd_inv_sqrt
    ZeroRankError
"""
from ._utils import drop_nan_rows, is_diagonal_matrix, is_square_matrix, is_sorted, allclose_sparse
from ._eigen import eigs, eig_corr, sort_eigs, spd_eig, spd_truncated_svd, \
    spd_inv, spd_inv_split, spd_inv_sqrt, ZeroRankError
from ._norm import schatten_norm



r"""This module provides implementations for the estimation package

.. moduleauthor:: B.Trendelkamp-Schroer <benjamin DOT trendelkamp-schroer AT fu-berlin DOT de>
"""

from . import count_matrix
from . import connectivity
from . import effective_counts
from . import likelihood
from . import transition_matrix
from . import prior

from .mle import mle_trev, mle_trev_given_pi



r"""
.. currentmodule: deeptime.markov

===============================================================================
Transition counting and analysis tools
===============================================================================

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    pcca
    PCCAModel

    TransitionCountEstimator
    TransitionCountModel

    ReactiveFlux
    reactive_flux

    MembershipsChapmanKolmogorovValidator

===============================================================================
Utilities
===============================================================================

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    number_of_states
    count_states
    compute_connected_sets

    compute_dtrajs_effective
    compute_effective_stride

    sample.indices_by_distribution
    sample.compute_index_states
    sample.indices_by_sequence
    sample.indices_by_state
"""

from ._util import number_of_states, count_states, compute_connected_sets, \
    compute_dtrajs_effective, compute_effective_stride
from . import sample
from . import tools  # former msmtools

from ._base import BayesianPosterior, _MSMBaseEstimator, MembershipsChapmanKolmogorovValidator
from ._pcca import pcca, PCCAModel
from ._transition_counting import TransitionCountEstimator, TransitionCountModel

from ._reactive_flux import ReactiveFlux, reactive_flux

from . import msm
from . import hmm



import numpy as np
from scipy.sparse import csr_matrix, diags
from scipy.sparse.linalg import LinearOperator

from . import objective_sparse_ops as ops


def convert_solution(z: np.ndarray, Cs):
    if not isinstance(Cs, csr_matrix):
        """Convert to csr_matrix"""
        Cs = csr_matrix(Cs)

    M = Cs.shape[0]
    x = z[0:M]
    y = z[M:]

    data = Cs.data
    indptr = Cs.indptr
    indices = Cs.indices

    data_P = np.zeros_like(data)
    diag_P = np.zeros(M, dtype=data.dtype)
    nu = np.zeros(M, dtype=data.dtype)

    ops.convertImpl(M, x, y, data, nu, data_P, diag_P, indices, indptr)

    P = csr_matrix((data_P, indices, indptr), shape=(M, M)) + diags(diag_P, 0)
    return nu / nu.sum(), P


def F(z: np.ndarray, Cs, c: np.ndarray):
    r"""Monotone mapping for the reversible MLE problem.

    Parameters
    ----------
    z : (2*M,) ndarray
        Point at which to evaluate mapping, z=(x, y)
    C : (M, M) scipy.sparse matrix
        Count matrix of reversible chain

    Returns
    -------
    Fval : (2*M,) ndarray
        Value of the mapping at z

    """
    if not isinstance(Cs, csr_matrix):
        """Convert to csr_matrix"""
        Cs = csr_matrix(Cs)

    M = Cs.shape[0]

    x = z[0:M]
    y = z[M:]

    data = Cs.data
    indptr = Cs.indptr
    indices = Cs.indices

    Fval = np.zeros(2 * M, dtype=data.dtype)

    ops.FImpl(M, x, y, c, data, Fval, indices, indptr)

    return Fval


def _DF(z: np.ndarray, Cs, c: np.ndarray):
    r"""Jacobian of the monotone mapping.

    Parameters
    ----------
    z : (2*M,) ndarray
        Point at which to evaluate mapping, z=(x, y)
    C : (M, M) scipy.sparse matrix
        Count matrix of reversible chain

    Returns
    -------
    DFval : (2*M, 2*M) scipy.sparse matrix
        Value of the Jacobian at z

    """
    if not isinstance(Cs, csr_matrix):
        """Convert to csr_matrix"""
        Cs = csr_matrix(Cs)

    M = Cs.shape[0]

    data = Cs.data
    indptr = Cs.indptr
    indices = Cs.indices

    """All subblocks DF_ij can be written as follows, DF_ij = H_ij +
    D_ij. H_ij has the same sparsity structure as C+C.T and D_ij is a
    diagonal matrix, i, j \in {x, y}
    """

    data_Hxx = np.zeros_like(data)
    data_Hyx = np.zeros_like(data)
    data_Hyy = np.zeros_like(data)

    diag_Dxx = np.zeros(M, dtype=data.dtype)
    diag_Dyx = np.zeros(M, dtype=data.dtype)
    diag_Dyy = np.zeros(M, dtype=data.dtype)

    x = z[0:M]
    y = z[M:]

    ops.DFImpl(M, x, y, data, data_Hxx, data_Hyy, data_Hyx, diag_Dxx, diag_Dyy, diag_Dyx, indices, indptr)

    Hxx = csr_matrix((data_Hxx, indices, indptr), shape=(M, M))
    Dxx = diags(diag_Dxx, 0)
    DFxx = Hxx + Dxx

    Hyy = csr_matrix((data_Hyy, indices, indptr), shape=(M, M))
    Dyy = diags(diag_Dyy, 0)
    DFyy = Hyy + Dyy

    Hyx = csr_matrix((data_Hyx, indices, indptr), shape=(M, M))
    Dyx = diags(diag_Dyx, 0)
    DFyx = Hyx + Dyx

    return DFxx, DFyx, DFyy


def DFsym(z: np.ndarray, Cs, c: np.ndarray):
    r"""Jacobian of the monotone mapping. Symmetric version.

    Parameters
    ----------
    z : (2*M,) ndarray
        Point at which to evaluate mapping, z=(x, y)
    C : (M, M) scipy.sparse matrix
        Count matrix of reversible chain

    Returns
    -------
    DFval : (2*M, 2*M) scipy.sparse matrix
        Value of the Jacobian at z

    """
    DFxx, DFyx, DFyy = _DF(z, Cs, c)
    return JacobianOperatorSymmetric(DFxx, DFyx, DFyy)


class JacobianOperatorSymmetric(LinearOperator):
    r"""Realise the following symmetric block sparse matrix.

           A   B.T
    M = (          )
           B   C

    with symmetric sub blocks A=A.T, C=C.T

    """

    def __init__(self, A, B, C):
        self.N1 = A.shape[0]
        self.N2 = C.shape[0]
        N = self.N1 + self.N2
        super().__init__(A.dtype, (N, N))

        self.A = A
        self.B = B
        self.BT = B.T.tocsr()
        self.C = C
        self.diag = np.hstack((A.diagonal(), C.diagonal()))

    def _matvec(self, x):
        N1 = self.N1
        N2 = self.N2
        y = np.zeros(N1 + N2, dtype=self.dtype)
        y[0:N1] = self.A.dot(x[0:N1]) + self.BT.dot(x[N1:])
        y[N1:] = self.B.dot(x[0:N1]) + self.C.dot(x[N1:])
        return y

    def diagonal(self):
        return self.diag



import warnings
from typing import Optional

import numpy as np
from scipy.sparse import issparse
from ..tools.estimation import effective_count_matrix
from .._base import _MSMBaseEstimator
from . import MarkovStateModel
from .. import TransitionCountEstimator, TransitionCountModel, count_states

from . import _koopman_reweighted_msm_impl as _impl
from ...util.matrix import submatrix
from ...util.types import ensure_dtraj_list

__author__ = 'Feliks Nüske, Fabian Paul, marscher, clonker'


class KoopmanReweightedMSM(MarkovStateModel):
    r""" This class belongs to a markov state model which was estimated by Koopman reweighting.

    Parameters
    ----------
    transition_matrix : (n, n) ndarray
        The transition matrix, see :meth:`MarkovStateModel.__init__` .
    stationary_distribution : ndarray(n), optional, default=None
        Stationary distribution if already computed, see :meth:`MarkovStateModel.__init__` .
    reversible : bool, optional, default=None
        Whether MSM is reversible, see :meth:`MarkovStateModel.__init__` .
    n_eigenvalues : int, optional, default=None
        The number of eigenvalues / eigenvectors to be kept, see :meth:`MarkovStateModel.__init__` .
    ncv : int optional, default=None
        Performance parameter for reversible MSMs, see :meth:`MarkovStateModel.__init__` .
    count_model : TransitionCountModel, optional, default=None
        In case the model was estimated with a :class:`OOMReweightedMSM` estimator, this contains a count matrix
        based on lagged data, i.e., data with the last :code:`lag` frames removed and histogram information
        based on the full dtrajs.
    twostep_count_matrices : (n, n, n) ndarray, optional, default=None
        Two-step count matrices for all states, each :code:`twostep_count_matrices[:, i, :]` is a count matrix
        for each :math:`i=1,\ldots,n`.
    oom_components : (m, n, m) ndarray, optional, default=None
        Matrix of set-observable operators, where :code:`m` is the rank based on the rank decision made during
        estimation.
    oom_eigenvalues : (m,) ndarray, optional, default=None
        The eigenvalues from OOM.
    oom_evaluator : (m,) ndarray, optional, default=None
        Evaluator of OOM.
    oom_information_state_vector : (m,) ndarray, optional, default=None
        Information state vector of OOM.

    See Also
    --------
    MaximumLikelihoodMSM
    BayesianMSM
    """

    def __init__(self, transition_matrix: np.ndarray, stationary_distribution: Optional[np.ndarray] = None,
                 reversible: Optional[bool] = None, n_eigenvalues: Optional[int] = None, ncv: Optional[int] = None,
                 twostep_count_matrices: Optional[np.ndarray] = None, oom_components: Optional[np.ndarray] = None,
                 count_model: Optional[TransitionCountModel] = None, oom_eigenvalues: Optional[np.ndarray] = None,
                 oom_evaluator: Optional[np.ndarray] = None, oom_information_state_vector: Optional[np.ndarray] = None):
        super(KoopmanReweightedMSM, self).__init__(
            transition_matrix, stationary_distribution=stationary_distribution, reversible=reversible,
            n_eigenvalues=n_eigenvalues, ncv=ncv, count_model=count_model
        )
        self._oom_eigenvalues = oom_eigenvalues
        self._oom_components = oom_components
        self._oom_information_state_vector = oom_information_state_vector
        self._oom_evaluator = oom_evaluator
        self._twostep_count_matrices = twostep_count_matrices
        if oom_evaluator is not None:
            self._oom_rank = oom_evaluator.size

    @property
    def twostep_count_matrices(self):
        r""" Two-step count matrices for all states. C2t[:, n, :] is a count matrix for each n. """
        return self._twostep_count_matrices

    @property
    def oom_eigenvalues(self):
        """System eigenvalues estimated by OOM."""
        return self._oom_eigenvalues

    @property
    def oom_timescales(self):
        """System timescales estimated by OOM."""
        return -self.lagtime / np.log(np.abs(self._oom_eigenvalues[1:]))

    @property
    def oom_rank(self):
        """Return OOM model rank."""
        return self._oom_rank

    @property
    def oom_components(self):
        """Return OOM components."""
        return self._oom_components

    @property
    def oom_information_state_vector(self):
        """ Return OOM initial state vector."""
        return self._oom_information_state_vector

    @property
    def oom_evaluator(self):
        """Return OOM evaluator vector."""
        return self._oom_evaluator


class OOMReweightedMSM(_MSMBaseEstimator):
    r"""OOM (observable operator model) MSM estimator for MSMs given discrete trajectory statistics.
    Here, each transition is re-weighted using OOM theory. Details can be found in :footcite:`nuske2017markov`.

    Parameters
    ----------
    lagtime : int
        lag time at which transitions are counted and the transition matrix is
        estimated.
    reversible : bool, optional, default = True
        If true compute reversible MSM, else non-reversible MSM
    count_mode : str, optional, default='sliding'
        mode to obtain count matrices from discrete trajectories. Should be
        one of:

        * 'sliding' : A trajectory of length T will have :math:`T-\tau` counts
          at time indexes

          .. math::

             (0 \rightarrow \tau), (1 \rightarrow \tau+1), ..., (T-\tau-1 \rightarrow T-1)
        * 'sample' : A trajectory of length T will have :math:`T/\tau` counts
          at time indexes

          .. math::

                (0 \rightarrow \tau), (\tau \rightarrow 2 \tau), ..., (((T/\tau)-1) \tau \rightarrow T)

    sparse : bool, optional, default = False
        If true compute count matrix, transition matrix and all derived
        quantities using sparse matrix algebra. In this case python sparse
        matrices will be returned by the corresponding functions instead of
        numpy arrays. This behavior is suggested for very large numbers of
        states (e.g. > 4000) because it is likely to be much more efficient.
    time_unit : str, optional, default='1 step'
        Description of the physical time of the input trajectories. May be used
        by analysis algorithms such as plotting tools to pretty-print the axes.
        By default '1 step', i.e. there is no physical time unit. Specify by a
        number, whitespace and unit. Permitted units are (* is an arbitrary
        string):

        |  'fs',  'femtosecond*'
        |  'ps',  'picosecond*'
        |  'ns',  'nanosecond*'
        |  'us',  'microsecond*'
        |  'ms',  'millisecond*'
        |  's',   'second*'

    nbs : int, optional, default=10000
        number of re-samplings for rank decision in OOM estimation.
    rank_mode : str, optional
        Re-sampling method for model rank selection. Can be

        * 'bootstrap_counts': Directly re-sample transitions based on effective count matrix.

        * 'bootstrap_trajs': Re-draw complete trajectories with replacement.

    tol_rank: float, optional, default = 10.0
        signal-to-noise threshold for rank decision.
    connectivity_threshold : float or '1/n'
        minimum number of counts to consider a connection between two states.
        Counts lower than that will count zero in the connectivity check and
        may thus separate the resulting transition matrix. The default
        evaluates to 1/n_states.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, lagtime, reversible=True, count_mode='sliding', sparse=False,
                 time_unit='1 step', nbs=10000, rank_mode='bootstrap_counts', tol_rank=10.0,
                 connectivity_threshold='1/n'):
        super(OOMReweightedMSM, self).__init__(reversible=reversible, sparse=sparse)
        # Check count mode:
        self.count_mode = str(count_mode).lower()
        if self.count_mode not in ('sliding', 'sample'):
            raise ValueError(f'count mode {count_mode} is unknown. Only \'sliding\' and \'sample\' are allowed.')
        if rank_mode not in ('bootstrap_counts', 'bootstrap_trajs'):
            raise ValueError('rank_mode must be either \'bootstrap_counts\' or \'bootstrap_trajs\'')
        self.nbs = nbs
        self.tol_rank = tol_rank
        self.rank_mode = rank_mode
        self.lagtime = lagtime
        self.time_unit = time_unit
        self.connectivity_threshold = connectivity_threshold

    def fetch_model(self) -> KoopmanReweightedMSM:
        r""" Yields the estimated Koopman reweighted MSM.

        Returns
        -------
        model : KoopmanReweightedMSM or None
            The estimated Koopman reweighted MSM or None if fit was not called yet.
        """
        return self._model

    def fit(self, dtrajs, **kw):
        r""" Fits an MSM using Koopman reweighting.

        Parameters
        ----------
        dtrajs : array_like or list of array_like
            Discrete trajectories.
        **kw
            API compatibility to sklearn, not actually used in algorithm.
        """
        dtrajs = ensure_dtraj_list(dtrajs)
        # remove last lag steps from dtrajs:
        dtrajs_lag = [traj[:-self.lagtime] for traj in dtrajs]

        # statistics are collected over full trajectories
        histogram = count_states(dtrajs, ignore_negative=True)
        # because of double counting, only count lagged trajs
        count_matrix = TransitionCountEstimator.count(count_mode=self.count_mode, dtrajs=dtrajs_lag,
                                                      lagtime=self.lagtime, sparse=self.sparse)
        count_model = TransitionCountModel(count_matrix, counting_mode=self.count_mode, lagtime=self.lagtime,
                                           state_histogram=histogram)
        count_model = count_model.submodel_largest(connectivity_threshold=self.connectivity_threshold, directed=True)

        # Estimate transition matrix using re-sampling:
        if self.rank_mode == 'bootstrap_counts':
            effective_count_mat = effective_count_matrix(dtrajs_lag, self.lagtime)
            Ceff = submatrix(effective_count_mat, count_model.state_symbols)
            smean, sdev = _impl.bootstrapping_count_matrix(Ceff, nbs=self.nbs)
        else:
            smean, sdev = _impl.bootstrapping_dtrajs(dtrajs_lag, self.lagtime, count_model.n_states_full, nbs=self.nbs,
                                                     active_set=count_model.state_symbols)
        # Estimate two step count matrices:
        twostep_count_matrices = _impl.twostep_count_matrix(dtrajs, self.lagtime, count_model.n_states_full)
        # Rank decision:
        rank_ind = _impl.rank_decision(smean, sdev, tol=self.tol_rank)
        # Estimate OOM components:
        if issparse(count_model.count_matrix_full):
            cmat = count_model.count_matrix_full.toarray()
        else:
            cmat = count_model.count_matrix_full
        oom_components, omega, sigma, eigenvalues = _impl.oom_components(
            cmat, twostep_count_matrices, rank_ind=rank_ind, lcc=count_model.state_symbols
        )
        # Compute transition matrix:
        P, lcc_new = _impl.equilibrium_transition_matrix(oom_components, omega, sigma, reversible=self.reversible)

        # Update active set and derived quantities:
        if lcc_new.size < count_model.n_states:
            assert isinstance(count_model, TransitionCountModel)
            count_model = count_model.submodel(count_model.symbols_to_states(lcc_new))
            warnings.warn("Caution: Re-estimation of count matrix resulted in reduction of the active set.")

        self._model = KoopmanReweightedMSM(
            transition_matrix=P,
            oom_eigenvalues=eigenvalues,
            oom_evaluator=sigma,
            oom_information_state_vector=omega,
            count_model=count_model,
            oom_components=oom_components,
            twostep_count_matrices=twostep_count_matrices
        )

        return self



r"""

==============
MSM estimation
==============

.. currentmodule:: deeptime.markov.tools.estimation

This module (:mod:`deeptime.markov.tools.estimation`) contains utility functions dealing with MSM estimation from data.

Countmatrix
===========

.. autosummary::
   :toctree: generated/

   count_matrix - estimate count matrix from discrete trajectories

Connectivity
============

.. autosummary::
   :toctree: generated/

   connected_sets - Find connected subsets
   largest_connected_set - Find largest connected set
   largest_connected_submatrix - Count matrix on largest connected set
   is_connected - Test count matrix connectivity

Estimation
==========

.. autosummary::
   :toctree: generated/

   transition_matrix - Estimate transition matrix
   rate_matrix
   log_likelihood
   tmatrix_cov
   error_perturbation


Sampling
========

.. autosummary::
   :toctree: generated/

   tmatrix_sampler - Random sample from transition matrix posterior

Bootstrap
=========

.. autosummary::
   :toctree: generated/

   bootstrap_counts
   bootstrap_trajectories

Priors
======

.. autosummary::
   :toctree: generated/

   prior_neighbor
   prior_const
   prior_rev


"""
from .api import *



from typing import Optional

import numpy as np
import scipy

from ..base import Estimator
from ..decomposition import TransferOperatorModel
from ..kernels import Kernel
from ..numeric import sort_eigs
from ..util.types import to_dataset


class KernelCCAModel(TransferOperatorModel):
    r""" The model produced by the :class:`KernelCCA` estimator.

    Parameters
    ----------
    eigenvalues : (n, ) ndarray
        The eigenvalues.
    eigenvectors : (m, n) ndarray
        The eigenvectors of the nonlinear transform of the input data.

    See Also
    --------
    KernelCCA
    """

    def __init__(self, data, kernel: Kernel, eigenvalues: np.ndarray, eigenvectors: np.ndarray):
        self._kernel = kernel
        self._eigenvalues = eigenvalues
        self._eigenvectors = eigenvectors
        self._data = data
        super().__init__(np.diag(np.sqrt(eigenvalues)),
                         instantaneous_obs=lambda x: (self.kernel.apply(x, self._data) @ self.eigenvectors),
                         timelagged_obs=lambda x: (self.kernel.apply(x, self._data) @ self.eigenvectors))

    @property
    def kernel(self) -> Kernel:
        r""" The kernel that was used for estimation. """
        return self._kernel

    @property
    def eigenvalues(self) -> np.ndarray:
        return self._eigenvalues

    @property
    def eigenvectors(self) -> np.ndarray:
        return self._eigenvectors


class KernelCCA(Estimator):
    r""" Estimator implementing the kernelized version of canonical correlation analysis.
    :footcite:`bach2002kernel` (CCA :footcite:`hotelling1992relations`)

    Parameters
    ----------
    kernel : Kernel
        The kernel to be used, see :mod:`deeptime.kernels` for a selection of predefined kernels.
    n_eigs : int
        Number of eigenvalue/eigenvector pairs to use for low-rank approximation.
    epsilon : float, optional, default=1e-6
        Regularization parameter.

    See Also
    --------
    KernelCCAModel

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, kernel: Kernel, n_eigs: int, epsilon: float = 1e-6):
        super().__init__()
        self.kernel = kernel
        self.n_eigs = n_eigs
        self.epsilon = epsilon

    def fit(self, data, **kwargs):
        r""" Fit this estimator instance onto data.

        Parameters
        ----------
        data
            Input data, see :meth:`to_dataset <deeptime.util.types.to_dataset>` for options.
        **kwargs
            Kwargs, may contain lagtime.

        Returns
        -------
        self : KernelCCA
            Reference to self.
        """
        dataset = to_dataset(data, lagtime=kwargs.get("lagtime", None))
        x, y = dataset[:]
        gram_0 = self.kernel.gram(x)
        gram_t = self.kernel.gram(y)

        # center Gram matrices
        n = x.shape[0]
        I = np.eye(n)  # identity
        N = I - np.full((n, n), fill_value=1. / n)  # centering matrix
        G_0 = np.linalg.multi_dot([N, gram_0, N])
        G_1 = np.linalg.multi_dot([N, gram_t, N])

        K = scipy.linalg.solve(G_0 + self.epsilon * I, G_0, assume_a='sym')
        Ak = scipy.linalg.solve(G_1 + self.epsilon * I, G_1, assume_a='sym')
        A = K @ Ak
        eigenvalues, eigenvectors = scipy.linalg.eig(A)
        eigenvalues, eigenvectors = sort_eigs(eigenvalues, eigenvectors)

        # determine effective rank m and perform low-rank approximations.
        if eigenvalues.shape[0] > self.n_eigs:
            eigenvectors = eigenvectors[:, :self.n_eigs]
            eigenvalues = eigenvalues[:self.n_eigs]

        self._model = KernelCCAModel(x, self.kernel, eigenvalues, eigenvectors)
        return self

    def fetch_model(self) -> Optional[KernelCCAModel]:
        r""" Yields the latest estimated model or None.

        Returns
        -------
        model : KernelCCAModel or None
            The latest estimated model or None.
        """
        return super().fetch_model()



def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('covar_c', parent_package, top_path)

    config.add_extension('_covartools',
                         sources=['covartools.cpp'],
                         include_dirs=['include'],
                         language='c++',
                         )
    return config
